
from functools import reduce

import numpy as np
from scipy.linalg import \
    det
from scipy.spatial.distance import \
    cdist
from traits.api import \
    Bool, Float, provides, \
    Int, Trait, List, Any, \
    Delegate, Property, cached_property, Dict, \
    Type, Array, HasTraits
import numpy as np

from .i_fets import IFETSEval
from bmcs_utils.api import InteractiveModel

#-------------------------------------------------------------------
# Numpy extension
#-------------------------------------------------------------------
def oriented_3d_array(arr, axis):
    '''In order to use the indices as spatial locators
    the array of gauss points is augmented with newaxes into 3D
    so that the cartesian broadcasting can be done.

    There is only the expand_dim function available in numpy.
    Here we want to  put the supplied array in 3d space along
    the axis  
    '''
    shape = [None, None, None]
    shape[axis] = slice(None)
    _arr = np.array(arr, dtype='float64')
    return _arr[tuple(shape)]

#-------------------------------------------------------------------
# FETSEval - general implementation of the fe-numerical quadrature
#-------------------------------------------------------------------


@provides(IFETSEval)
class FETSEval(InteractiveModel):

    dof_r = Array(np.float64,
                  desc='Local coordinates of nodes included in the field ansatz')

    geo_r = Array(np.float64,
                  desc='Local coordinates of nodes included in the geometry ansatz')

    n_nodal_dofs = Int(desc='Number of nodal degrees of freedom')

    id_number = Int

    #-------------------------------------------------------------------------
    # Derived info about the finite element formulation
    #-------------------------------------------------------------------------
    n_dof_r = Int

    def _n_dof_r_default(self):
        return len(self.dof_r)

    n_geo_r = Int

    def _n_geo_r_default(self):
        return len(self.geo_r)

    #-------------------------------------------------------------------------
    # Field visualization
    #-------------------------------------------------------------------------

    vtk_r = Array(
        np.float64, desc='Local coordinates of nodes included in the field visualization')

    vtk_cell_types = Any(
        desc='Tuple of vtk cell types in the same order as they are specified in the vtk_cells list')

    vtk_cells = List(
        desc='List of maps of nodes constituting the vtk cells covering the single element')

    vtk_cell_class = Property

    def _get_vtk_cell_class(self):
        from tvtk.tvtk_classes import tvtk_helper
        return tvtk_helper.get_class(self.vtk_cell_type)

    # Distinguish the type of base geometric entity to be used for
    # the visualization of the results.
    #
    # field_entity_type = Enum('vertex','line','triangle','quad','tetra','hexa')

    vtk_node_cell_data = Property(depends_on='vtk_cells, vtk_cell_types')

    @cached_property
    def _get_vtk_node_cell_data(self):

        n_cells = len(self.vtk_cells)

        # check if vtk_cell_types is a list, if not make one
        if isinstance(self.vtk_cell_types, str):
            cell_classes = [self.vtk_cell_types for i in range(n_cells)]
        else:
            cell_classes = self.vtk_cell_types

        from tvtk.tvtk_classes import tvtk_helper

        cell_types = []
        for cell_str in cell_classes:
            cell_class = tvtk_helper.get_class(cell_str)
            cell_types.append(cell_class().cell_type)

        if isinstance(self.vtk_cells[0], int):
            # just a single cell defined
            return (np.array([0, ], dtype=int),
                    np.array(self.vtk_cells.shape[0], dtype=int),
                    np.array(self.vtk_cells, dtype=int),
                    cell_types)

        offset_list = []
        length_list = []
        cell_list = []
        vtk_offset = 0
        for cell in self.vtk_cells:
            cell_len = len(cell)
            cell_list += cell
            length_list.append(cell_len)
            offset_list.append(vtk_offset)
            vtk_offset += cell_len + 1

        return (np.array(offset_list, dtype=int),
                np.array(length_list, dtype=int),
                np.array(cell_list, dtype=int),
                np.array(cell_types, dtype=int))

    vtk_ip_cell_data = Property(depends_on='vtk_cells, vtk_cell_types')

    @cached_property
    def _get_vtk_ip_cell_data(self):
        from tvtk.tvtk_classes import tvtk_helper

        n_ip_pnts = self.ip_coords.shape[0]

        cell_types = np.array([(tvtk_helper.get_class('PolyVertex')()).cell_type])

        return (np.array([0, ], dtype=int),
                np.array([n_ip_pnts], dtype=int),
                np.arange(n_ip_pnts),
                cell_types)

    n_vtk_r = Property(Int, depends_on='vtk_r')

    @cached_property
    def _get_n_vtk_r(self):
        return self.vtk_r.shape[0]

    n_vtk_cells = Property(Int, depends_on='field_faces')

    @cached_property
    def _get_n_vtk_cells(self):
        return self.field_faces.shape[0]

    vtk_pnt_ip_map = Property(Array(Int))

    @cached_property
    def _get_vtk_pnt_ip_map(self):
        return self.get_vtk_pnt_ip_map_data(self.vtk_r)

    def adjust_spatial_context_for_point(self, sctx):
        '''
        Method gets called prior to the evaluation at the material point level.

        The method can be used for dimensionally reduced evaluators.
        This is FETS specific and should be moved there.
        However, the RTraceEval is not distinguished at the moment, therefore
        it is here - move!!!.   
        '''
        sctx.X_reg = sctx.X

    def get_vtk_pnt_ip_map_data(self, vtk_r):
        '''
        mapping of the visualization point to the integration points
        according to mutual proximity in the local coordinates
        '''
        vtk_pt_arr = np.zeros((1, 3), dtype='float64')
        ip_map = np.zeros(vtk_r.shape[0], dtype='int_')
        for i, vtk_pt in enumerate(vtk_r):
            vtk_pt_arr[0, self.dim_slice] = vtk_pt[self.dim_slice]
            # get the nearest ip_coord
            ip_map[i] = np.argmin(cdist(vtk_pt_arr, self.ip_coords))
        return np.array(ip_map)

    #-------------------------------------------------------------------------
    # NUMERICAL INTEGRATION
    #-------------------------------------------------------------------------
    #
    # The integration grid is constructed using broadcasting
    # provided by numpy. This provides a loopless implementation
    # with well defined ordering of gauss points and further the
    # slicing of gauss points using the numpy indexing slices.
    #
    # The old loop-based expansion is preserved below for reference.
    #
    gp_r_grid = Property(depends_on='ngp_r,ngp_s,ngp_t')

    @cached_property
    def _get_gp_r_grid(self):
        '''Return a tuple of three arrays for X, Y, Z coordinates of the
        gauss points within the element.
        '''
        # get the oriented arrays for each direction prepared for broadcasting
        #
        gp_coords = [oriented_3d_array(self._GP_COORDS[ngp], dim)
                     for dim, ngp
                     in enumerate(self.n_gp_list)]

        # broadcast the values to construct all combinations of all gauss point
        # coordinates.
        #
        x, y, z = np.broadcast_arrays(*gp_coords)
        return x, y, z

    gp_w_grid = Property(depends_on='ngp_r,ngp_s,ngp_t')

    @cached_property
    def _get_gp_w_grid(self):
        '''In analogy to the above, get the grid of gauss weights in 3D.
        '''
        # get the oriented arrays for each direction prepared for broadcasting
        #
        gp_w = [oriented_3d_array(self._GP_WEIGHTS[ngp], dim) for dim, ngp
                in enumerate(self.n_gp_list)]

        # broadcast the values to construct all combinations of all gauss point
        # coordinates.
        #
        w = reduce(lambda x, y: x * y, gp_w)
        return w

    ip_coords = Property(depends_on='ngp_r,ngp_s,ngp_t')

    def _get_ip_coords(self):
        '''Generate the flat array of ip_coords used for integration.
        '''
        x, y, z = self.gp_r_grid
        return np.c_[x.flatten(), y.flatten(), z.flatten()]

    r_m = Property

    def _get_r_m(self):
        return self.ip_coords

    ip_coords_grid = Property(depends_on='ngp_r,ngp_s,ngp_t')

    def _get_ip_coords_grid(self):
        '''Generate the grid of ip_coords
        '''
        return np.c_[self.gp_r_grid]

    ip_weights = Property(depends_on='ngp_r,ngp_s,ngp_t')

    w_m = Property

    def _get_w_m(self):
        return self.ip_weights

    n_m = Property

    def _get_n_m(self):
        return len(self.ip_weights)

    def _get_ip_weights(self):
        '''Generate the flat array of ip_coords used for integration.
        '''
        w = self.gp_w_grid
        return w.flatten()

    ip_weights_grid = Property(depends_on='ngp_r,ngp_s,ngp_t')

    def _get_ip_weights_grid(self):
        '''Generate the flat array of ip_coords used for integration.
        '''
        w = self.gp_w_grid
        return w

    def get_ip_scheme(self, *params):
        return (self.ip_coords, self.ip_weights)

    n_gp = Property(depends_on='ngp_r,ngp_s,ngp_t')

    @cached_property
    def _get_n_gp(self):
        nr = max(1, self.ngp_r)
        ns = max(1, self.ngp_s)
        nt = max(1, self.ngp_t)
        return nr * ns * nt

    n_gp_list = Property(depends_on='ngp_r,ngp_s,ngp_r')

    @cached_property
    def _get_n_gp_list(self):
        nr = self.ngp_r
        ns = self.ngp_s
        nt = self.ngp_t
        return [nr, ns, nt]
    #-------------------------------------------------------------------------
    # SUBSPACE INTEGRATION
    #-------------------------------------------------------------------------
    #
    # Get the integration scheme for a subspace specified using the slicing indexes.
    #

    def get_sliced_ip_scheme(self, ip_idx_list):

        minmax = {0: min, -1: max}
        w = []
        r = []
        ix = []
        for dim_idx, ip_idx in enumerate(ip_idx_list):
            if isinstance(ip_idx, int):
                w.append(1.0)
                r.append(minmax[ip_idx](self.dof_r[:, dim_idx]))
            elif isinstance(ip_idx, slice):
                # if there is a slice - put the array in the corresponding dimension
                # - only the full slide - i.e. slice(None,None,None)
                #   is allowed
                n_gp = self.n_gp_list[dim_idx]
                w.append(oriented_3d_array(self._GP_WEIGHTS[n_gp], dim_idx))
                r.append(oriented_3d_array(self._GP_COORDS[n_gp], dim_idx))
                ix.append(dim_idx)
        r_grid = np.broadcast_arrays(*r)
        r_c = np.c_[tuple([r.flatten() for r in r_grid])]
        w_grid = reduce(lambda x, y: x * y, w)
        if isinstance(w_grid, float):
            w_grid = np.array([w_grid], dtype='float64')
        else:
            w_grid = w_grid.flatten()
        return r_c, w_grid, ix

    #-------------------------------------------------------------------------
    # FIELD TRACING / VISUALIZATION
    #-------------------------------------------------------------------------
    # The user-specified fv_loc_coords list gets transform to an internal
    # array representation
    #
    vtk_r_arr = Property(depends_on='vtk_r')

    @cached_property
    def _get_vtk_r_arr(self):
        if len(self.vtk_r) == 0:
            raise ValueError(
                'Cannot generate plot, no vtk_r specified in fets_eval')
        return np.array(self.vtk_r)

    def get_vtk_r_glb_arr(self, X_mtx, r_mtx=None):
        '''
        Get an array with global coordinates of the element decomposition.

        If the local_point_list is non-empty then use it instead of the one supplied 
        by the element specification. This is useful for augmented specification of RTraceEval 
        evaluations with a specific profile of a field variable to be traced.
        '''
        if self.dim_slice:
            X_mtx = X_mtx[:, self.dim_slice]

        if r_mtx == None:
            r_mtx = self.vtk_r_arr

        # TODO - efficiency in the extraction of the global coordinates. Is broadcasting
        # a possibility - we only need to augment the matrix with zero coordinates in the
        # in the unhandled dimensions.
        #
        X3D = np.array([np.dot(self.get_N_geo_mtx(r_pnt), X_mtx)[0, :]
                     for r_pnt in r_mtx])
        n_dims = r_mtx.shape[1]
        n_add = 3 - n_dims
        if n_add > 0:
            X3D = np.hstack([X3D,
                          np.zeros([r_mtx.shape[0], n_add], dtype='float64')])
        return X3D

    def get_X_pnt(self, sctx):
        '''
        Get the global coordinates for the specified local coordinats r_pnt
        @param r_pnt: local coordinates
        '''
        r_pnt = sctx.r_pnt
        X_mtx = sctx.X

        return np.einsum('', self.Nr_i_geo, X_mtx)
        return dot(self.get_N_geo(r_pnt), X_mtx)

    def get_x_pnt(self, sctx):
        '''
        Get the global coordinates for the specified local coordinats r_pnt
        @param r_pnt: local coordinates
        '''
        r_pnt = sctx.r_pnt
        x_mtx = sctx.x

        # TODO - efficiency in the extraction of the global coordinates. Is broadcasting
        # a possibility - we only need to augment the matrix with zero coordinates in the
        # in the unhandled dimensions.
        #
        return np.dot(self.get_N_geo(r_pnt), x_mtx)

    def map_r2X(self, r_pnt, X_mtx):
        '''
        Map the local coords to global
        @param r_pnt: local coords
        @param X_mtx: matrix of the global coords of geo nodes
        '''
        # print "mapping ",dot( self.get_N_geo_mtx(r_pnt)[0], X_mtx )," ",
        # r_pnt
        return np.dot(self.get_N_geo(r_pnt), X_mtx)

    # Number of element DOFs
    #
    n_e_dofs = Int

    # Dimensionality
    dim_slice = None

    # Parameters for the time-loop
    #
    def new_cntl_var(self):
        return np.zeros(self.n_e_dofs, np.float64)

    def new_resp_var(self):
        return np.zeros(self.n_e_dofs, np.float64)

    def get_state_array_size(self):
        r_range = max(1, self.ngp_r)
        s_range = max(1, self.ngp_s)
        t_range = max(1, self.ngp_t)
        return self.m_arr_size * r_range * s_range * t_range

    m_arr_size = Property()

    @cached_property
    def _get_m_arr_size(self):
        return self.get_mp_state_array_size(None)

    ngp_r = Int(0, label='Number of Gauss points in r-direction')
    ngp_s = Int(0, label='Number of Gauss points in s-direction')
    ngp_t = Int(0, label='Number of Gauss points in t-direction')

    # if no gauss point is defined in one direction (e.g. for ngp_t=0 for a 2D-problem)
    # then the default value for ngp_t=0 is used and a weighting coefficient of value 1.
    # In 'get_gp' the maximum of npg and 1 is used as range in the loop which leads to
    # a simple multiplication with 1 for that direction. The coordinate of the gauss point
    # in this direction is set to zero ('gp' : [0.]).

    # The index of the entry corresponds to the order of the polynomial
    #
    _GP_COORDS = [[0.],  # for polynomial order  = 0
                  [0.],  # for polynomial order  = 1
                  # por polynomial order = 2
                  [-0.57735026918962584, 0.57735026918962584],
                  # for polynomial order = 3
                  [-0.7745966692414834, 0., 0.7745966692414834],
                  # for polynomial order = 4
                  [-0.861136311594053, -0.339981043584856,
                   0.339981043584856, 0.861136311594053]
                  ]

    _GP_WEIGHTS = [[1.],
                   [2.],
                   [1., 1.],
                   [0.55555555555555558, 0.88888888888888884,
                       0.55555555555555558],
                   [0.347854845137454, 0.652145154862546,
                       0.652145154862546, 0.347854845137454]
                   ]

    #-------------------------------------------------------------------------
    # Epsilon as an engineering value
    #-------------------------------------------------------------------------
    def get_eps0_eng(self, sctx, u):
        '''Get epsilon without the initial strain
        '''
        if self.mats_eval.initial_strain:
            X_pnt = self.get_X_pnt(sctx)
            x_pnt = self.get_x_pnt(sctx)
            eps0_mtx = self.mats_eval.initial_strain(X_pnt, x_pnt)
            return self.mats_eval.map_eps_mtx_to_eng(eps0_mtx)
        else:
            return None

    def get_eps_eng(self, sctx, u):
        X_mtx = sctx.X
        r_pnt = sctx.loc
        B_mtx = self.get_B_mtx(r_pnt, X_mtx)
        eps_eng = np.dot(B_mtx, u)
        return eps_eng

    def get_eps1t_eng(self, sctx, u):
        '''Get epsilon without the initial strain
        '''
        eps_eng = self.get_eps_eng(sctx, u)
        eps0_eng = self.get_eps0_eng(sctx, u)
        if eps0_eng != None:
            eps_eng -= eps0_eng
        return eps_eng

    #-------------------------------------------------------------------------
    # Epsilon as a tensorial value - used for visualization in mayavi
    #-------------------------------------------------------------------------
    def get_eps0_mtx33(self, sctx, u):
        '''Get epsilon without the initial strain
        '''
        eps0_mtx33 = np.zeros((3, 3), dtype='float64')
        if self.mats_eval.initial_strain:
            X_pnt = self.get_X_pnt(sctx)
            x_pnt = self.get_x_pnt(sctx)
            eps0_mtx33[self.dim_slice, self.dim_slice] = self.mats_eval.initial_strain(
                X_pnt, x_pnt)
        return eps0_mtx33

    def get_eps_mtx33(self, sctx, u):
        eps_eng = self.get_eps_eng(sctx, u)
        eps_mtx33 = np.zeros((3, 3), dtype='float64')
        eps_mtx33[self.dim_slice, self.dim_slice] = self.mats_eval.map_eps_eng_to_mtx(
            eps_eng)
        return eps_mtx33

    def get_eps1t_mtx33(self, sctx, u):
        '''Get epsilon without the initial strain
        '''
        eps_mtx33 = self.get_eps_mtx33(sctx, u)
        eps0_mtx33 = self.get_eps0_mtx33(sctx, u)
        return eps_mtx33 - eps0_mtx33

    #-------------------------------------------------------------------------
    # Displacement in a n arbitrary point within an element
    #-------------------------------------------------------------------------
    def map_eps(self, sctx, u):
        X_mtx = sctx.X
        r_pnt = sctx.loc
        B_mtx = self.get_B_mtx(r_pnt, X_mtx)
        eps = np.dot(B_mtx, u)
        return eps

    def get_u(self, sctx, u):
        N_mtx = self.get_N_mtx(sctx.loc)
        return np.dot(N_mtx, u)




from numpy import array, where, append
from traits.api import \
    HasTraits, Int, Array, Property, cached_property, List, Trait, Dict, \
    Any


class SysMtxArray(HasTraits):
    '''Class managing an array of equally sized matrices with 
    enumerated equations.

    It is used as an intermediate result of finite element integration
    of the system matrix. From this format, it is converted to a sparse
    matrix representation. For example, the coord format, the conversion 
    is possible by flattening the value array using the ravel.
    (see e.g. coo_mtx) 
    '''
    dof_map_arr = Array('int_')
    mtx_arr = Array('float64')

    n_dofs = Property(depends_on='dof_map_arr')

    @cached_property
    def _get_n_dofs(self):
        if self.dof_map_arr.shape[0] == 0:
            return 0
        return self.dof_map_arr.max() + 1

    def get_dof_ix_array(self, dof):
        '''Return the element number and index of the dof within the element
        '''
        return where(self.dof_map_arr == dof)

    def _zero_rc(self, dof_ix_array):
        '''Set row column values associated with dof a
        to zero.  
        '''
        el_arr, row_arr = dof_ix_array
        for el, i_dof in zip(el_arr, row_arr):
            self.mtx_arr[el, i_dof, :] = 0.0
            self.mtx_arr[el, :, i_dof] = 0.0

    def _add_col_to_vector(self, dof_ix_array, F, factor):
        '''Get the slice of the a-th column.
        (used for the implementation of the essential boundary conditions)
        '''
        el_arr, row_arr = dof_ix_array
        for el, i_dof in zip(el_arr, row_arr):
            rows = self.dof_map_arr[el]
            F[rows] += factor * self.mtx_arr[el, :, i_dof]

    def _get_diag_elem(self, dof_ix_array):
        '''Get the value of diagonal element at a-ths dof. 
        '''
        K_aa = 0.
        el_arr, row_arr = dof_ix_array
        for el, i_dof in zip(el_arr, row_arr):
            K_aa += self.mtx_arr[el, i_dof, i_dof]
        return K_aa

    def _add_diag_elem(self, dof_ix_array, K_aa):
        '''Get the value of diagonal element at a-ths dof. 
        '''
        el_arr, row_arr = dof_ix_array
        el = el_arr[0]
        i_dof = row_arr[0]
        self.mtx_arr[el, i_dof, i_dof] = K_aa

    def _get_col_subvector(self, dof_ix_array):
        idx_arr = array([], dtype=int)
        val_arr = array([], dtype=float)
        el_arr, row_arr = dof_ix_array
        for el, i_dof in zip(el_arr, row_arr):
            r_dofs = self.dof_map_arr[el]
            idx_arr = append(idx_arr, r_dofs)
            val_arr = append(val_arr, self.mtx_arr[el, :, i_dof])

        return idx_arr, val_arr



'''
'''
from math import \
    pi

from ibvpy.api import TLoop, TLine
from ibvpy.tmodel.mats2D.mats2D_cmdm.mats2D_cmdm import MATS2DMicroplaneDamage
from ibvpy.tmodel.mats2D.mats2D_explorer_bcond import get_value_and_coeff, BCDofProportional
from ibvpy.mathkit.mfn import MFnLineArray

import numpy as np
import pylab as p

if __name__ == '__main__':

    elastic_debug = True
    # Tseval for a material model
    #
    tseval = MATS2DMicroplaneDamage(elastic_debug=elastic_debug)

    value, coeff = get_value_and_coeff(1., 0.0)

#    bcond_alpha = BCDof(var='u', dof=0, value=value,
    bcond_alpha = BCDofProportional(var='u', dof=0, value=value,
                     link_dofs=[1],
                     link_coeffs=[coeff],
                     time_function=lambda t: t)

    ts = TStepper(tse=tseval,
             bcond_list=[ bcond_alpha
                         ],
             rtrace_list=[ RTDofGraph(name='strain 0 - stress 0',
                                  var_x='eps_app', idx_x=0,
                                  var_y='sig_app', idx_y=0,
                                  record_on='update'),
                         RTDofGraph(name='strain 1 - stress 1',
                                  var_x='eps_app', idx_x=1,
                                  var_y='sig_app', idx_y=1,
                                  record_on='update'),
                         RTDofGraph(name='strain 0 - stress 1',
                                  var_x='eps_app', idx_x=0,
                                  var_y='sig_app', idx_y=1,
                                  record_on='update'),
                         RTDofGraph(name='strain 1 - stress 0',
                                  var_x='eps_app', idx_x=1,
                                  var_y='sig_app', idx_y=0,
                                  record_on='update'),
                         RTDofGraph(name='strain 0 - strain 1',
                                  var_x='eps_app', idx_x=0,
                                  var_y='eps_app', idx_y=1,
                                  record_on='update'),
                         ]
                         )

    # Put the time-stepper into the time-loop
    #
    if elastic_debug:
        tmax = 1.
        n_steps = 1
    else:
        tmax = 0.001
        # tmax = 0.0006
        n_steps = 100

    tl = TLoop(tstepper=ts,
             DT=tmax / n_steps, KMAX=100, RESETMAX=0,
             min=0.0, max=tmax)
#             T=TRange(min=0.0, max=tmax))

    from numpy import argmax

    alpha_arr = np.linspace(-pi / 2 * 1.05, 2 * (pi / 2.) + pi / 2.*0.05, 20)

    sig0_m_list = []
    sig1_m_list = []
    eps0_m_list = []
    eps1_m_list = []

    for alpha in alpha_arr:

        value, coeff = get_value_and_coeff(1., alpha)
        bcond_alpha.value = value
        bcond_alpha.link_coeffs[0] = coeff

        tl.eval()

        eps0_sig0 = tl.rtrace_mngr.rtrace_bound_list[0]
        eps1_sig1 = tl.rtrace_mngr.rtrace_bound_list[1]
#        eps0_sig0 = tl.rv_mngr.rv_list[0]
#        eps1_sig1 = tl.rv_mngr.rv_list[1]

        sig0_midx = argmax(np.fabs(eps0_sig0.trace.ydata))
        sig1_midx = argmax(np.fabs(eps1_sig1.trace.ydata))

        sig0_m = eps0_sig0.trace.ydata[ sig0_midx ]
        sig1_m = eps1_sig1.trace.ydata[ sig1_midx ]

        eps0_m = eps0_sig0.trace.xdata[ sig0_midx ]
        eps1_m = eps1_sig1.trace.xdata[ sig1_midx ]

        sig0_m_list.append(sig0_m)
        sig1_m_list.append(sig1_m)
        eps0_m_list.append(eps0_m)
        eps1_m_list.append(eps1_m)

    sig_plot = MFnLineArray(xdata=sig0_m_list,
                              ydata=sig1_m_list)
    eps_plot = MFnLineArray(xdata=eps0_m_list,
                              ydata=eps1_m_list)

#    sig_plot.configure_traits()

    print('sig_plot.xdata', sig_plot.ydata)
    p.plot(sig_plot.xdata, sig_plot.ydata)
    p.show()

    # Put the time-loop into the simulation-framework and map the
    # object to the user interface.
    #




from traits.api import \
    HasTraits, WeakRef, Tuple, Property, cached_property, Any


class CellGridSlice(HasTraits):
    '''General implementation of a slice within the FEGrid
    '''
    cell_grid = WeakRef()  # 'ibvpy.mesh.cell_grid.cell_grid.CellGrid' )

    grid_slice = Any

    idx_tuple = Property(
        Tuple, depends_on='grid_slice,cell_grid.+changed_structure')

    @cached_property
    def _get_idx_tuple(self):
        ''' Classify the index specification
            By default, the idx is assigned to element index
            to handle the case of a single integer index
            '''
        idx1 = self.grid_slice
        # The default for node index is the universal slice
        idx2 = slice(None, None)
        n_dims = self.cell_grid.n_dims
        # If the index is an iterable
        if hasattr(self.grid_slice, '__iter__'):
            # Get the first n_dims of indexes for the cell grid
            idx1 = self.grid_slice[:n_dims]
            # If there are more indexes than n_dims save them for
            # use within the cell to identify the nodes.
            if len(self.grid_slice) > n_dims:
                idx2 = self.grid_slice[n_dims:]
        return (idx1, idx2)

    elem_grid = Property

    def _get_elem_grid(self):
        idx1, _ = self.idx_tuple
        return self.cell_grid.cell_idx_grid[idx1]

    elems = Property

    def _get_elems(self):
        # get the cells affected by the slice @todo - rename cell_grid
        # attribute within CellGrid
        return self.elem_grid.flatten()

    # Get the node map associated with the sliced elements
    #
    nodes = Property

    def _get_nodes(self):
        # get the node map associated with the sliced elements
        idx1, idx2 = self.idx_tuple

        sliced_cell_node_map = self.cell_grid.cell_node_map[self.elems]

        # get the sliced nodes for the sliced cells
        return sliced_cell_node_map[:, self.cell_grid.grid_cell[idx2]]

    points = Property

    def _get_points(self):
        # get the coordinates for the sliced coords
        return self.cell_grid.point_X_arr[self.nodes]

    # Global coordinates of nodes involved in the slice
    # Structured element by element
    #
    point_X_arr = Property

    def _get_point_X_arr(self):
        return self.cell_grid.point_X_arr[self.nodes]

    # Parametric coordinates of nodes involved in the slice
    # Structured element by element
    #
    point_x_arr = Property

    def _get_point_x_arr(self):
        return self.cell_grid.point_x_arr[self.nodes]


if __name__ == '__main__':

    from .cell_grid import CellGrid
    from .cell_spec import CellSpec

    cell_grid = CellGrid(shape=(2, 8),
                         grid_cell_spec=CellSpec(node_coords=[[-1, -1],
                                                              [1, -1],
                                                              [0, 0],
                                                              [1, 1],
                                                              [-1, 1]]))

    print('cell_grid')
    print(cell_grid.cell_idx_grid)

    print('cell_grid points')
    print(cell_grid.point_idx_grid)

    print('grid_cell')
    print(cell_grid.grid_cell)

    print('grid_cell[:,-1]')
    print(cell_grid.grid_cell[:, -1])

    print('sliced cells')
    cell_grid_slice = cell_grid[:, -1, :, -1]

    print('elemes')
    print(cell_grid_slice.elems)

    print('nodes')
    print(cell_grid_slice.nodes)

    print('global coords')
    print(cell_grid_slice.point_X_arr)

    print('local coords')
    print(cell_grid_slice.point_x_arr)

    print('cell grid')
    print(cell_grid_slice.cell_grid)

    print('get a single element - the first one - cell grid[0,0]')
    print(cell_grid[0].elems)

    print('get all nodes of the boundary elements in second-direction')
    print(cell_grid[:, -1, ...].nodes)



'''
Module providing subsidary functions for tensor mapping in 2D
'''

from numpy import \
     array, zeros, ix_, dot, hstack, vstack

from scipy.linalg import \
    inv


#---------------------------------------------------------------------------------------------
# Switch from engineering to tensorial notation 
#---------------------------------------------------------------------------------------------

def map2d_eps_eng_to_mtx( eps_eng ):
    '''
    Switch from engineering notation to tensor notation for strains in 2D
    '''        
    eps_mtx = array( [[eps_eng[0]     , eps_eng[2] / 2.],
                      [eps_eng[2] / 2., eps_eng[1]     ]] )
    return eps_mtx


def map2d_sig_eng_to_mtx( sig_eng ):
    '''
    Switch from engineering notation to tensor notation for stresses in 2D
    '''        
    sig_mtx = array( [[sig_eng[0] , sig_eng[2] ],
                      [sig_eng[2] , sig_eng[1] ]] )
    return sig_mtx


#---------------------------------------------------------------------------------------------
# Switch from tensorial to engineering notation  
#---------------------------------------------------------------------------------------------

def map2d_eps_mtx_to_eng( eps_mtx ):
    '''
    Switch from tensor notation to engineering notation strains in 2D
    '''       
    eps_eng = array([ eps_mtx[0,0], eps_mtx[1,1], 2.0 * eps_mtx[0,1] ])
    return eps_eng


def map2d_sig_mtx_to_eng( sig_mtx ):
    '''
    Switch from tensor notation to engineering notation for stresses in 2D
    '''        
    sig_eng = array([ sig_mtx[0,0], sig_mtx[1,1], sig_mtx[0,1] ])
    return sig_eng


#---------------------------------------------------------------------------------------------
# Subsidiary index mapping functions for rank-four to rank-two tensors
#---------------------------------------------------------------------------------------------

def map2d_ijkl2mn( i,j,k,l ):
    '''
    Map the four-rank indexes to the two-rank matrix using the major
    and minor symmetry.
    '''
    # 2D-case:
    # first two indices (ij)
    if i==0 and j==0:
        m = 0
    elif i==1 and j==1:
        m = 1
    elif (i==0 and j==1) or (i==1 and j==0):
        m = 2
        
    # second two indices (kl)
    if k==0 and l==0:
        n = 0
    elif k==1 and l==1:
        n = 1
    elif (k==0 and l==1) or (k==1 and l==0):
        n = 2
        
    return m,n


#---------------------------------------------------------------------------------------------
# Subsidiary mapping functions for rank-two to rank-four tensor 
#---------------------------------------------------------------------------------------------

def map2d_tns2_to_tns4( tns2 ):
    '''
    Map a matrix to a fourth order tensor assuming minor and major symmetry,
    e.g. D_mtx (3x3) in engineering notation to D_tns(2,2,2,2)).
    '''
    n_dim = 2
    tns4 = zeros([n_dim,n_dim,n_dim,n_dim])
    for i in range(0,n_dim):
        for j in range(0,n_dim):
            for k in range(0,n_dim):
                for l in range(0,n_dim):
                    tns4[i,j,k,l] = tns2[map2d_ijkl2mn(i,j,k,l)]
    return tns4

               
#---------------------------------------------------------------------------------------------
# Subsidiary mapping functions for rank-four to rank-two tensor 
#---------------------------------------------------------------------------------------------

def map2d_tns4_to_tns2( tns4 ):
    '''
    Map a fourth order tensor to a matrix assuming minor and major symmetry,
    e.g. D_tns(2,2,2,2) to D_mtx (3x3) in engineering notation.
    (Note: Explicit assignment of components used for speedup.)
    '''
    n_eng = 3
    tns2 = zeros([n_eng,n_eng])
    
    tns2[0,0] =             tns4[0,0,0,0]
    tns2[0,1] = tns2[1,0] = tns4[0,0,1,1]
    tns2[0,2] = tns2[2,0] = tns4[0,0,0,1]
    
    tns2[1,1] =             tns4[1,1,1,1]
    tns2[1,2] = tns2[2,1] = tns4[1,1,0,1]
    
    tns2[2,2] =             tns4[0,1,0,1]
    
    return tns2

#----------------------------------------------------------------------------------------------
# Compliance mapping 2D (used for inversion of the damage effect tensor in engineering notation 
#----------------------------------------------------------------------------------------------

def compliance_mapping2d( C_mtx_2d ):
    '''
    The components of the compliance matrix are multiplied with factor 1,2 or 4 depending on their 
    position in the matrix (due to symmetry and switching of tensorial to engineering notation
    (Note: gamma_xy = 2*epsilon_xy, etc.). Necessary for evaluation of D=inv(C). 
    '''
    idx1 = [0,1]
    idx2 = [2]
    C11 = C_mtx_2d[ix_(idx1,idx1)]
    C12 = C_mtx_2d[ix_(idx1,idx2)]
    C21 = C_mtx_2d[ix_(idx2,idx1)]
    C22 = C_mtx_2d[ix_(idx2,idx2)]
    return vstack( [ hstack( [C11,   2*C12] ),
                     hstack( [2*C21, 4*C22] ) ] )

#----------------------------------------------------------------------------------------------
# Reduce the 3D-elasticity/compliance matrix to the 2D-cases plane strain or plane stress 
#----------------------------------------------------------------------------------------------

def get_D_plane_stress( D_mtx_3d ):
    '''
    Reduce the 6x6-elasticity matrix for the 3D-case to a 
    3x3 matrix for the 2D case assuming plane stress (sig_yz=0, sig_zz=0, sig_xz=0)
    '''
    idx2 = [0,1,5]
    idx3 = [2,3,4]
    D22 = D_mtx_3d[ix_(idx2,idx2)]
    D23 = D_mtx_3d[ix_(idx2,idx3)]
    D32 = D_mtx_3d[ix_(idx3,idx2)]
    D33 = D_mtx_3d[ix_(idx3,idx3)]
    D_pstress_term = dot( dot( D23, inv( D33 ) ), D32 )
    D_mtx_2d = D22 - D_pstress_term
    return D_mtx_2d


def get_D_plane_strain( D_mtx_3d ):
    '''
    Reduce the 6x6-elasticity matrix for the 3D-case to a 
    3x3 matrix for the 2D case assuming plane strain (eps_yz=0, eps_zz=0, eps_xz=0)
    '''
    idx2 = [0,1,5]
    D_mtx_2d = D_mtx_3d[ix_(idx2,idx2)]
    return D_mtx_2d


def get_C_plane_stress( C_mtx_3d ):
    '''
    Reduce the 6x6-compliance matrix for the 3D-case to a 
    3x3 matrix for the 2D case assuming plane stress (sig_yz=0, sig_zz=0, sig_xz=0)
    '''
    idx2 = [0,1,5]
    C_mtx_2d = C_mtx_3d[ix_(idx2,idx2)]
    return C_mtx_2d


def get_C_plane_strain( C_mtx_3d ):
    '''
    Reduce the 6x6-compliance matrix for the 3D-case to a 
    3x3 matrix for the 2D case assuming plane strain (eps_yz=0, eps_zz=0, eps_xz=0)
    '''
    idx2 = [0,1,5]
    idx3 = [2,3,4]
    C22 = C_mtx_3d[ix_(idx2,idx2)]
    C23 = C_mtx_3d[ix_(idx2,idx3)]
    C32 = C_mtx_3d[ix_(idx3,idx2)]
    C33 = C_mtx_3d[ix_(idx3,idx3)]
    C_pstrain_term = dot( dot( C23, inv( C33 ) ), C32 )
    C_mtx_2d = C22 - C_pstrain_term
    return C_mtx_2d

if __name__ == '__main__':
    from numpy import array
    tn2 = array([[1,4,5],
                 [4,2,6],
                 [5,6,3]])
    print('tn2')
    print(tn2)

    tn4 = map2d_tns2_to_tns4(tn2)
    
    print('tns2_to_tns4')
    print(tn4)
    
    tn2a = map2d_tns4_to_tns2(tn4)
    
    print('back')
    print('tns4_to_tns2')
    print(tn2a)
    



from numpy import linalg, ix_, zeros
from traits.api import HasTraits, Property, Any


class DenseMtx(HasTraits):
    '''Dense matrix with the interface of a sparse matrix assembly.
    Used for debugging and performance comparison of sparse solvers.
    '''
    assemb = Any

    mtx = Property

    def _get_mtx(self):
        n_dofs = self.assemb.n_dofs
        sys_mtx = zeros([n_dofs, n_dofs], dtype=float)
        # loop over the list of matrix arrays
        for sys_mtx_arr in self.assemb.get_sys_mtx_arrays():
            # loop over the matrix array
            for dof_map, mtx in zip(sys_mtx_arr.dof_map_arr,
                                    sys_mtx_arr.mtx_arr):
                sys_mtx[ix_(dof_map, dof_map)] += mtx
        return sys_mtx

    def solve(self, rhs):
        u_vct = linalg.solve(self.mtx, rhs)
        return u_vct

    def __str__(self):
        '''String representation - delegate to matrix'''
        return str(self.mtx)

    def __repr__(self):
        return repr(self.mtx)




from .i_bcond import IBCond
from traits.api import \
    Instance,  \
    List
from ibvpy.view.ui import BMCSTreeNode


class BCondMngr(BMCSTreeNode):

    node_name = 'boundary conditions'

    bcond_list = List(IBCond)

    def _tree_node_list_default(self):
        return self.bcond_list

    selected_bcond = Instance(IBCond)

    def setup(self, sctx):
        '''
        '''
        for bc in self.bcond_list:
            bc.setup(sctx)

    def register(self, K):
        '''Register the boundary condition in the equation system.
        '''
        for bcond in self.bcond_list:
            bcond.register(K)

    def apply(self, step_flag, sctx, K, R, t_n, t_n1):

        for bcond in self.bcond_list:
            bcond.apply(step_flag, sctx, K, R, t_n, t_n1)




'''
Created on 14.06.2017
@author: Yingxiong
'''
import mayavi.mlab as mlab
import numpy as np


min_sig = -20.0
max_sig = 5.0
n_sig = 100j
sig_1, sig_2, sig_3 = np.mgrid[min_sig: max_sig: n_sig,
                               min_sig: max_sig: n_sig,
                               min_sig: max_sig: n_sig]

sig_abcj = np.einsum('jabc->abcj', np.array([sig_1, sig_2, sig_3]))
DELTA = np.identity(3)
sig_abcij = np.einsum('abcj,jl->abcjl', sig_abcj, DELTA)

I1 = np.einsum('...ii,...ii', sig_abcij, DELTA)
s_ij = sig_abcij - np.einsum('...,ij->...ij', I1 / 3.0, DELTA)
J2 = np.einsum('...ii,...ii', s_ij, s_ij) / 2.0

k = 2.
f = J2 - k ** 2

f_pipe = mlab.contour3d(
    sig_1, sig_2, sig_3, f, contours=[0.0], color=(0, 1, 0))

# deviatoric_plane
pi_plane = mlab.pipeline.scalar_cut_plane(f_pipe, plane_orientation='z_axes')
pi_plane.implicit_plane.normal = (1., 1., 1.)
pi_plane.implicit_plane.widget.draw_plane = True
pi_plane.enable_contours = True
pi_plane.contour.minimum_contour = 0.0
pi_plane.contour.maximum_contour = 0.0
pi_plane.contour.number_of_contours = 1
pi_plane.actor.property.representation = 'points'

# plane stress
z_plane = mlab.pipeline.scalar_cut_plane(f_pipe, plane_orientation='z_axes')
z_plane.implicit_plane.origin = (0., 0., 0.)
z_plane.implicit_plane.widget.draw_plane = True
z_plane.enable_contours = True
z_plane.contour.minimum_contour = 0.0
z_plane.contour.maximum_contour = 0.0
z_plane.contour.number_of_contours = 1
z_plane.actor.property.representation = 'points'

# hydrostatic section
h_plane = mlab.pipeline.scalar_cut_plane(f_pipe, plane_orientation='z_axes')
h_plane.implicit_plane.normal = (1., -1., 0.)
h_plane.implicit_plane.widget.draw_plane = True
h_plane.enable_contours = True
h_plane.contour.minimum_contour = 0.0
h_plane.contour.maximum_contour = 0.0
h_plane.contour.number_of_contours = 1
h_plane.actor.property.representation = 'points'

mlab.axes(f_pipe)
# mlab.outline()
mlab.show()


from numpy import repeat, arange
from traits.api import \
    HasTraits, provides, \
    List, \
    Property, cached_property

from .i_fe_parent_domain import IFEParentDomain


@provides(IFEParentDomain)
class FEGridActivationMap(HasTraits):

    #-------------------------------------------------------------------------
    # Implement the parent interface
    #-------------------------------------------------------------------------

    # @todo: separate piece of functionality - does not use any parent-child link
    # should be captured in a separate piece of code.
    inactive_elems = List(changed_structure=True)

    def deactivate(self, idx_tuple):
        '''Exclude the specified element from the integration.
        '''
        offset = self.get_cell_offset(idx_tuple)
        self.inactive_elems.append(offset)

    def reactivate(self, idx_tuple):
        '''Include the element in the computation agaoin
        '''
        raise NotImplementedError

    # get boolean array with inactive elements indicated by False
    activation_map = Property(depends_on='inactive_elems,shape')

    @cached_property
    def _get_activation_map(self):
        amap = repeat(True, self.n_grid_elems)
        amap[self.inactive_elems] = False
        return amap

    # get indices of all active elements
    idx_active_elems = Property(depends_on='inactive_elems,shape')

    @cached_property
    def _get_idx_active_elems(self):
        return arange(self.n_grid_elems)[self.activation_map]

    n_active_elems = Property

    def _get_n_active_elems(self):
        if self.inactive_elems != None:
            return self.idx_active_elems.shape[0]
        else:
            return self.shape



'''
Created on Feb 11, 2018

@author: rch
'''

import os

from ibvpy.tmodel.viz3d_field import Vis3DField, Viz3DField
from mayavi import mlab
from mayavi.sources.vtk_xml_file_reader import VTKXMLFileReader
from tvtk.api import write_data

import numpy as np
import traits.api as tr


class Vis3DStateField(Vis3DField):

    mats = tr.WeakRef

    def update(self):
        ts = self.tstep
        U = ts.U_k
        t_n1 = ts.t_n1
        # this needs to be adopted to tstep[domain_state]
        # loop over the subdomains
        U_vector_fields = []
        state_fields = []
        for domain in ts.fe_domain:
            xdomain = domain.xmodel
            fets = xdomain.fets
            var_function = domain.tmodel.var_dict.get(self.var, None)
            if var_function == None or xdomain.hidden:
                continue
            eps_Enab = xdomain.map_U_to_field(U)
            state_field = var_function(eps_Enab, t_n1, **domain.state_k)
            #
            # state_field = domain.state_n.get(self.var, None)
            # if (xdomain.hidden) or (state_field is None):
            #     # If the state variable not present in the domain, skip
            #     continue
            state_fields.append(state_field.flatten())
            DELTA_x_ab = fets.vtk_expand_operator
            U_Eia = U[xdomain.o_Eia]
            _, _, n_a = U_Eia.shape
            U_vector_fields.append(np.einsum(
                'Ia,ab->Ib', U_Eia.reshape(-1, n_a), DELTA_x_ab
            ))
        if len(U_vector_fields) == 0:
            raise ValueError('no fields for variable %s' % self.var)
        self.ug.point_data.vectors = np.vstack(U_vector_fields)
        self.ug.point_data.vectors.name = 'displacement'
        self.ug.point_data.scalars = np.hstack(state_fields)
        self.ug.point_data.scalars.name = self.var
        fname = '%s_step_%008.4f' % (self.var, t_n1)
        target_file = os.path.join(
            self.dir, fname.replace('.', '_')
        ) + '.' + self.extension
        write_data(self.ug, target_file)
        self.add_file(target_file)


class Viz3DScalarField(Viz3DField):

    warp_factor = tr.Float(1.0, auto_set=False, enter_set=True)

    lut_manager = tr.Property

    def _get_lut_manager(self):
        lut = self.warp_vector.children[0]
        return lut.scalar_lut_manager

    def setup(self):
        m = mlab
        fname = self.vis3d.file_list[0]
        var = self.vis3d.var
        self.d = VTKXMLFileReader()
        self.d.initialize(fname)
        self.src = m.pipeline.add_dataset(self.d)
        self.warp_vector = m.pipeline.warp_vector(self.src)
        self.warp_vector.filter.scale_factor = self.warp_factor
        self.surf = m.pipeline.surface(self.warp_vector)
        lut = self.warp_vector.children[0]
        lut.scalar_lut_manager.trait_set(
            lut_mode='Reds',
            show_scalar_bar=True,
            show_legend=True,
            data_name=var,
            use_default_range=False,
            data_range=np.array([0, 1], dtype=np.float64),
        )

        lut.scalar_lut_manager.scalar_bar.width = 0.5
        lut.scalar_lut_manager.scalar_bar.height = 0.15
        lut.scalar_lut_manager.scalar_bar.orientation = 'horizontal'
        lut.scalar_lut_manager.scalar_bar_representation.trait_set(
            maximum_size=np.array([100000, 100000]),
            minimum_size=np.array([1, 1]),
            position=np.array([0.5, 0.05]),
            position2=np.array([0.45, 0.1]),
        )
        lut.scalar_lut_manager.label_text_property.trait_set(
            font_family='times',
            italic=False,
            bold=False
        )
        lut.scalar_lut_manager.title_text_property.trait_set(
            font_family='times',
            italic=False,
            bold=False
        )



'''
Created on Jun 18, 2009

@author: jakub
'''
from numpy import array, dot
from scipy.linalg import \
    inv
from traits.api import \
    Int, Array
from ibvpy.fets.fets_eval import FETSEval


#-----------------------------------------------------------------------------
# FEBar1D
#-----------------------------------------------------------------------------
class FETS1D2L3U(FETSEval):
    '''
    Fe Bar 3 nodes, deformation
    '''

    debug_on = True

    # Dimensional mapping
    dim_slice = slice(0, 1)

    n_e_dofs = Int(3)
    n_nodal_dofs = Int(1)

    dof_r = Array(value=[[-1], [0], [1]])
    geo_r = Array(value=[[-1], [1]])

    vtk_r = Array(value=[[-1.], [0.], [1.]])
    vtk_cells = [[0, 2, 1]]
    vtk_cell_types = 'QuadraticEdge'

    # Integration parameters
    #
    ngp_r = 3

    def get_N_geo_mtx(self, r_pnt):
        '''
        Return geometric shape functions
        @param r_pnt:
        '''
        r = r_pnt[0]
        N_mtx = array([[0.5 - r / 2., 0.5 + r / 2.]])
        return N_mtx

    def get_dNr_geo_mtx(self, r_pnt):
        '''
        Return the matrix of shape function derivatives.
        Used for the conrcution of the Jacobi matrix.

        @TODO - the B matrix is used
        just for uniaxial bar here with a trivial differential
        operator.
        '''
        return array([[-1. / 2, 1. / 2]])

    def get_N_mtx(self, r_pnt):
        '''
        Return shape functions
        @param r_pnt:local coordinates
        '''
        r = r_pnt[0]
        return array([[1. / 2 * r * (r - 1.), 1. - r * r, 1. / 2 * r * (r + 1.)]])

    def get_dNr_mtx(self, r_pnt):
        '''
        Return the derivatives of the shape functions
        '''
        r = r_pnt[0]
        return array([[r - 1. / 2., -2. * r, r + 1. / 2.]])

    def get_B_mtx(self, r, X):
        '''
        Return kinematic matrix
        @param r:local coordinates
        @param X:global coordinates
        '''
        J_mtx = self.get_J_mtx(r, X)
        dNr_mtx = self.get_dNr_mtx(r)
        B_mtx = dot(inv(J_mtx), dNr_mtx)
        return B_mtx

#----------------------- example --------------------


def example_with_new_domain():
    from ibvpy.api import \
        TStepper as TS, RTDofGraph, RTraceDomainListField, TLoop, \
        TLine, BCDof, IBVPSolve as IS, DOTSEval
    from ibvpy.tmodel.mats1D.mats1D_elastic.mats1D_elastic import MATS1DElastic

    fets_eval = FETS1D2L3U(mats_eval=MATS1DElastic(E=10.))

    from ibvpy.mesh.fe_grid import FEGrid

    # Discretization
    domain = FEGrid(coord_max=(3., ),
                    shape=(3, ),
                    fets_eval=fets_eval)

    ts = TS(dof_resultants=True,
            sdomain=domain,
            # conversion to list (square brackets) is only necessary for slicing of
            # single dofs, e.g "get_left_dofs()[0,1]"
            #         bcond_list =  [ BCDof(var='u', dof = 0, value = 0.)     ] +
            #                    [ BCDof(var='u', dof = 2, value = 0.001 ) ]+
            #                    [ )     ],
            bcond_list=[BCDof(var='u', dof=0, value=0.),
                        #                        BCDof(var='u', dof = 1, link_dofs = [2], link_coeffs = [0.5],
                        #                              value = 0. ),
                        #                        BCDof(var='u', dof = 2, link_dofs = [3], link_coeffs = [1.],
                        #                              value = 0. ),
                        BCDof(var='f', dof=6, value=1,
                              # link_dofs = [2], link_coeffs = [2]
                              )],
            rtrace_list=[RTDofGraph(name='Fi,right over u_right (iteration)',
                                    var_y='F_int', idx_y=0,
                                    var_x='U_k', idx_x=1),
                         RTraceDomainListField(name='Stress',
                                               var='sig_app', idx=0),
                         RTraceDomainListField(name='Displacement',
                                               var='u', idx=0),
                         RTraceDomainListField(name='N0',
                                               var='N_mtx', idx=0,
                                               record_on='update')

                         ]
            )

    # Add the time-loop control
    tloop = TLoop(tstepper=ts,
                  tline=TLine(min=0.0, step=1, max=1.0))

    print('---- result ----')
    print(tloop.eval())
    print(ts.F_int)
    print(ts.rtrace_list[0].trace.ydata)

    # Put the whole stuff into the simulation-framework to map the
    # individual pieces of definition into the user interface.
    #
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    app = IBVPyApp(ibv_resource=tloop)
    app.main()


if __name__ == '__main__':
    example_with_new_domain()




#from sys_matrix import SysSparseMtx, SysDenseMtx
import unittest

from numpy import array, zeros, arange, array_equal, hstack, dot
from numpy import array, zeros, arange, array_equal, sqrt
from scipy.linalg import solve, norm

from .coo_mtx import COOSparseMtx
from .dense_mtx import DenseMtx
from .sys_mtx_assembly import SysMtxAssembly


# bar clamped at the left end and loaded at the right end
def get_bar_mtx_array(shape, k=10):
    '''Get an array of matrices and dof_maps corresponding to
    a bar discretization
    '''
    el_mtx = array([[k, -k],
                    [-k, k]], dtype='float64')

    el_mtx_arr = array([el_mtx for i in range(shape)], dtype=float)
    el_dof_map = array([arange(shape),
                        arange(shape) + 1], dtype=int).transpose()

    return el_dof_map, el_mtx_arr


class TestSysMtxSolver(unittest.TestCase):
    '''
    Test the matrix functionality required for Kling the
    system matrix, construction of the sparsity map and solution 
    of the system.

    The sparse matrices are constructed using the matrix Kly 
    list.
    '''

    def setUp(self):
        '''
        Set up a simple 3x3 matrix corresponding to a tension bar
        connected in the middle node.

        The first DOF is factored prescribed to be zero - 
        the off-diagonal terms are therefore deleted 
        (only 1 remains at the diagonal)
        '''
        n_dofs = 3
        shape = 2
        el_mtx = array([[1, -1],
                        [-1, 1]], dtype='float64')

        el_mtx_arr = array([el_mtx for i in range(shape)], dtype=float)
        el_dof_map = array([arange(n_dofs - 1),
                            arange(n_dofs - 1) + 1], dtype=int).transpose()

        # Kly object
        self.sys_K = SysMtxAssembly()

        self.sys_K.add_mtx_array(dof_map_arr=el_dof_map,
                                 mtx_arr=el_mtx_arr)
        self.sys_K.register_constraint(0)
        self.rhs = zeros(n_dofs)
        self.rhs[-1] = 1.

        # solve the system using standard scipy methods
        # on a full matrix
        #
        la_mtx = array([[1, 0, 0],
                        [0, 2, -1],
                        [0, -1, 1]], dtype=float)
        la_rhs = array([0, 0, 1], dtype=float)

        self.la_u = solve(la_mtx, la_rhs)

    def test_coo_sparse_mtx(self):
        '''Construct the coordinate sparsity map and matrix and solve it.
        '''
        u = self.sys_K.solve(self.rhs, matrix_type='coord')
        self.assertTrue(array_equal(u, self.la_u))

    def test_dense_mtx(self):
        '''Construct the dense matrix and solve it.
        '''
        u = self.sys_K.solve(self.rhs, matrix_type='dense')
        self.assertTrue(array_equal(u, self.la_u))


class TestSysMtxConstraints(unittest.TestCase):
    '''
    Test functionality connected with the application of
    constraints.
    '''

    def test_bar1(self):
        '''Clamped bar loaded at the right end with unit displacement
        [00]-[01]-[02]-[03]-[04]-[05]-[06]-[07]-[08]-[09]-[10]
        'u[0] = 0, u[10] = 1'''
        K = SysMtxAssembly()
        dof_map, mtx_arr = get_bar_mtx_array(shape=10)
        K.add_mtx_array(dof_map_arr=dof_map, mtx_arr=mtx_arr)
        K.register_constraint(a=0,  u_a=0.)  # clamped end
        K.register_constraint(a=10, u_a=1.)
        K.register_constraint(a=10, u_a=1.)
        # add load
        R = zeros(K.n_dofs)
        # system solver
        u = K.solve(R)
        # expected solution
        u_ex = array([0., 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.],
                     dtype=float)
        difference = sqrt(norm(u - u_ex))
        self.assertAlmostEqual(difference, 0)

    def test_bar2(self):
        '''Clamped bar composed of two linked bars loaded at the right end
        [00]-[01]-[02]-[03]-[04]-[05]-[06]-[07]-[08]-[09]-[10]
        [11]-[12]-[13]-[14]-[15]-[16]-[17]-[18]-[19]-[20]-[21]
        u[0] = 0, u[5] = u[16], R[-1] = R[21] = 10
        '''
        K = SysMtxAssembly()
        dof_map1, mtx_arr1 = get_bar_mtx_array(shape=10)
        K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
        dof_map2, mtx_arr2 = get_bar_mtx_array(shape=10)
        # Note the dof map of the second br must be adjusted to start
        # the DOF enumeration with 3
        n_dofs1 = K.n_dofs
        K.add_mtx_array(dof_map_arr=dof_map2 + n_dofs1, mtx_arr=mtx_arr2)
        # add constraints
        K.register_constraint(a=0, u_a=0.)  # clamped end
        K.register_constraint(a=5, alpha=[1], ix_a=[16])
        # add load
        R = zeros(K.n_dofs)
        R[-1] = 10
        # system solver
        u = K.solve(R)
        # expected solution
        u_ex = array([0., 1., 2., 3., 4., 5., 5., 5., 5., 5., 5.,
                      5., 5., 5., 5., 5., 5., 6., 7., 8., 9., 10.],
                     dtype=float)
        difference = sqrt(norm(u[1] - u_ex[1]))
        self.assertAlmostEqual(difference, 0)

        #
        # '---------------------------------------------------------------'
        # 'Clamped bar composed of two linked bars control displ at right'
        # 'u[0] = 0, u[5] = u[16], u[21] = 1'
        # Remove the load and put a unit displacement at the right end
        # Note, the load is irrelevant in this case and will be rewritten
        #
        K.register_constraint(a=21, u_a=1)
        # system solver
        u = K.solve(R, matrix_type='dense')
        # expected solution
        u_ex = array([0.,   1 / 10.,  2 / 10., 3 / 10., 4 / 10., 5 / 10., 5 / 10., 5 / 10.,
                      5 / 10., 5 / 10.,  5 / 10.,  5 /
                      10.,  5 / 10.,  5 / 10.,  5 / 10.,
                      5 / 10.,  5 / 10.,  6 / 10.,  7 / 10.,  8 / 10.,  9 / 10.,  1.],
                     dtype=float)
        difference = sqrt(norm(u[-1] - u_ex[-1]))
        self.assertAlmostEqual(difference, 0)
        #

    def test_bar3(self):
        '''Clamped bar with recursive constraints (load at right end)
        [0]-[1]-[2]-[3]
        u[1] = 0.2 * u[2], u[2] = 0.2 * u[3], R[3] = 10
        '''
        K = SysMtxAssembly()
        dof_map, mtx_arr = get_bar_mtx_array(shape=3)
        K.add_mtx_array(dof_map_arr=dof_map, mtx_arr=mtx_arr)
        K.register_constraint(a=0, u_a=0.)  # clamped end
        K.register_constraint(a=1, alpha=[0.5], ix_a=[2])
        K.register_constraint(a=2, alpha=[1], ix_a=[3])
        # add load
        R = zeros(K.n_dofs)
        R[3] = 1
        # system solver
        # K.apply_constraints(R)
        u = K.solve(R)
        # expected solution
        u_ex = array([-0.,  0.1, 0.2, 0.2],
                     dtype=float)
        difference = sqrt(norm(u - u_ex))
        self.assertAlmostEqual(difference, 0)
        #
        # '---------------------------------------------------------------'
        # 'Clamped bar with recursive constraints (displ at right end)'
        # 'u[1] = 0.5 * u[2], u[2] = 1.0 * u[3], u[3] = 1'
        K.register_constraint(a=3, u_a=1)
        # system solver
        u = K.solve(R)
        # expected solution
        u_ex = array([0.,  0.5, 1,  1], dtype=float)
        difference = sqrt(norm(u - u_ex))
        self.assertAlmostEqual(difference, 0)
        #

    def test_bar4(self):
        '''Clamped bar 3 domains, each with 2 elems (displ at right end)
        [0]-[1]-[2] [3]-[4]-[5] [6]-[7]-[8]
        u[0] = 0, u[2] = u[3], u[5] = u[6], u[8] = 1'''
        K = SysMtxAssembly()
        dof_map1, mtx_arr1 = get_bar_mtx_array(shape=2)
        K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
        dof_map2, mtx_arr2 = get_bar_mtx_array(shape=2)
        K.add_mtx_array(dof_map_arr=dof_map2 + 3, mtx_arr=mtx_arr2)
        dof_map3, mtx_arr3 = get_bar_mtx_array(shape=2)
        K.add_mtx_array(dof_map_arr=dof_map3 + 6, mtx_arr=mtx_arr3)
        # add constraints
        K.register_constraint(a=0, u_a=0.)  # clamped end
        K.register_constraint(a=2, alpha=[1], ix_a=[3])
        K.register_constraint(a=5, alpha=[1], ix_a=[6])
        K.register_constraint(a=8, u_a=1.)  # loaded end
        # add load
        R = zeros(K.n_dofs)
        # system solver
        u = K.solve(R)
        # expected solution
        u_ex = array([0., 1 / 6.,  1 / 3., 1 / 3., 1 / 2., 2 / 3., 2 / 3.,  5 / 6.,  1.],
                     dtype=float)
        difference = sqrt(norm(u - u_ex))
        self.assertAlmostEqual(difference, 0)

    def test_bar5(self):
        '''Clamped bar with 4 elements. Elements 2-4 are reinforced
        with another bar with 3 elements
        [0]-[1]-[2]-[3]-[4]
            [5]-[6]-[7]'''
        # 'u[0] = 0, u[1] = u[5], u[3] = u[7], u[4] = 1'
        K = SysMtxAssembly()
        dof_map1, mtx_arr1 = get_bar_mtx_array(shape=4)
        K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
        dof_map2, mtx_arr2 = get_bar_mtx_array(shape=2)
        K.add_mtx_array(dof_map_arr=dof_map2 + 5, mtx_arr=mtx_arr2)
        # add constraints
        K.register_constraint(a=0, u_a=0.)  # clamped end
        K.register_constraint(a=1, alpha=[1], ix_a=[5])
        K.register_constraint(a=3, alpha=[1], ix_a=[7])
        K.register_constraint(a=4, u_a=1.)  # loaded end
        # add load
        R = zeros(K.n_dofs)
        # system solver
        u = K.solve(R)
        # expected solution
        u_ex = array([0., 1 / 3.,  0.5,
                      2 / 3.,  1.,
                      1 / 3.,  0.5,         2 / 3.],
                     dtype=float)
        difference = sqrt(norm(u - u_ex))
        self.assertAlmostEqual(difference, 0)
        #

    def test_bar6(self):
        '''Clamped bar with 4 elements. Elements 2-4 are reinforced 
        with another bar with 1 element linked proportianally
        [0]-[1]-[2]-[3]-[4]
              [5]-[6]
        u[0] = 0, u[1] = u[5], u[3] = u[7], u[4] = 1'''
        K = SysMtxAssembly()
        dof_map1, mtx_arr1 = get_bar_mtx_array(shape=4)
        K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
        dof_map2, mtx_arr2 = get_bar_mtx_array(shape=1)
        K.add_mtx_array(dof_map_arr=dof_map2 + 5, mtx_arr=mtx_arr2)
        # add constraints
        K.register_constraint(a=0, u_a=0.)  # clamped end
        K.register_constraint(a=5, alpha=[0.5, 0.5], ix_a=[1, 2])
        K.register_constraint(a=6, alpha=[0.5, 0.5], ix_a=[2, 3])
        K.register_constraint(a=4, u_a=1.)  # loaded end
        # add load
        R = zeros(K.n_dofs)
        # system solver
        u = K.solve(R)
        # expected solution
        u_ex = array([-0.,  0.3, 0.5, 0.7, 1.,  0.4,  0.6], dtype=float)
        difference = sqrt(norm(u - u_ex))
        self.assertAlmostEqual(difference, 0)

    def test_bar7(self):
        '''Two clamped beams link in parallel
        and loaded by force at right end
        [5]-[6]-[7]-[8]-[9]
        [0]-[1]-[2]-[3]-[4]
        u[5] = u[0], u[0] = 0, u[4] = u[9], R[4] = 1'''
        K = SysMtxAssembly()
        dof_map1, mtx_arr1 = get_bar_mtx_array(shape=4)
        K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
        dof_map2, mtx_arr2 = get_bar_mtx_array(shape=4)
        K.add_mtx_array(dof_map_arr=dof_map2 + 5, mtx_arr=mtx_arr2)
        # add constraints
        K.register_constraint(a=5, alpha=[1], ix_a=[0])
        K.register_constraint(a=0)  # clamped end
        K.register_constraint(a=4, alpha=[0.5], ix_a=[9])
        # add load
        R = zeros(K.n_dofs)
        # load at the coupled end nodes is doubled
        R[9] = 1
        R[4] = 1
        # system solver
#        print 'K\n',K
#        print 'R\n',R
        u = K.solve(R)
#        print 'K\n',K
#        print 'R\n',R
#        print 'u\n',u
        # expected solution
        u_ex = array([-0., 0.06, 0.12,  0.18,  0.24,  0.,
                      0.12,  0.24, 0.36,  0.48],
                     dtype=float)
        difference = sqrt(norm(u - u_ex))
        self.assertAlmostEqual(difference, 0)

    def test_bar8(self):
        '''Overruling of a non-zero constraint by a zero-constraint
        applied for corner and edge constrains.
        [0]-[1]
        u[0] = 0, u[1] = 0, u[1] = 4'''
        K = SysMtxAssembly()
        dof_map1, mtx_arr1 = get_bar_mtx_array(shape=1)
        K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
        # add constraints
        K.register_constraint(a=0, u_a=0)  # clamped end
        K.register_constraint(a=1, u_a=4)  # should be ignored
        K.register_constraint(a=1, u_a=0)  # should remain
        # add load
        R = zeros(K.n_dofs)
        # load at the coupled end nodes is doubled
        # system solver
        u = K.solve(R)
        # expected solution
        u_ex = array([-0., -0.], dtype=float)
        self.assertAlmostEqual(u[1], u_ex[1])

    def test_bar9(self):
        '''Zero terms  at the diagonal. combined with zero-value constraints
        (simulating the deactivation of elements)
        [0]-[1]
        u[0] = 0, u[1] = 0'''
        K = SysMtxAssembly()

        dof_map1, mtx_arr1 = get_bar_mtx_array(shape=1, k=0.)
        K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
        # add constraints
        K.register_constraint(a=0, u_a=0)  # clamped end
        K.register_constraint(a=1, u_a=0)  # clamped end
        # add load
        R = zeros(K.n_dofs)
        R[1] = 10.  # irrelevant
        # load at the coupled end nodes is doubled
        # system solver
        u = K.solve(R)
        # expected solution
        u_ex = array([-0., -0.], dtype=float)
        self.assertAlmostEqual(u[1], u_ex[1])

    def test_bar10(self):
        '''Zero terms  at the diagonal. combined with zero-value constraints
        (simulating the deactivation of elements)
        [0]-[1]-[2]-[3]-[4]  [5]-[6]
        u[0] = 0, u[4] = u[5]  u[6] = 0, K[5,6] = 0, R[3] = 1'''
        K = SysMtxAssembly()
        dof_map1, mtx_arr1 = get_bar_mtx_array(shape=4)
        K.add_mtx_array(dof_map_arr=dof_map1, mtx_arr=mtx_arr1)
        dof_map2, mtx_arr2 = get_bar_mtx_array(shape=1, k=0.)
        K.add_mtx_array(dof_map_arr=dof_map2 + 5, mtx_arr=mtx_arr2)
        # add constraints
        K.register_constraint(a=6, u_a=0.)
        K.register_constraint(a=0)  # clamped end
        K.register_constraint(a=4, alpha=[1], ix_a=[5])
        # add load
        R = zeros(K.n_dofs)
        # load at the coupled end nodes is doubled
        R[3] = 1
        # system solver
        u = K.solve(R)
        # expected solution
        u_ex = array([-0., 0.1, 0.2, 0.3, 0.3, 0.3, 0.],
                     dtype=float)
        for uu, ue in zip(u, u_ex):
            self.assertAlmostEqual(uu, ue)

    def test_bar11(self):
        '''Zero terms  at the diagonal. combined with zero-value constraints
        (simulating the deactivation of elements)
        [0]-[1]-[2]-[3]-[4]  [5]-[6]
        u[0] = 0, u[4] = u[5]  u[6] = 0, K[5,6] = 0, R[3] = 1'''
        K = SysMtxAssembly()
        dof_map2, mtx_arr2 = get_bar_mtx_array(shape=1, k=0)
        K.add_mtx_array(dof_map_arr=dof_map2, mtx_arr=mtx_arr2)
        dof_map1, mtx_arr1 = get_bar_mtx_array(shape=4)
        K.add_mtx_array(dof_map_arr=dof_map1 + 2, mtx_arr=mtx_arr1)
        # add constraints
        K.register_constraint(a=6, u_a=0.)
        K.register_constraint(a=2, alpha=[1], ix_a=[1])
        K.register_constraint(a=0)  # clamped end
        # add load
        R = zeros(K.n_dofs)
        # load at the coupled end nodes is doubled
        R[3] = - 1
        # system solver
        u = K.solve(R)
        # expected solution
        u_ex = array([-0.,  -0.3, -0.3, -0.3, -0.2, -0.1, -0.],
                     dtype=float)
        for uu, ue in zip(u, u_ex):
            self.assertAlmostEqual(uu, ue)



#-------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Aug 19, 2009 by: rch, ascholzen

from math import sqrt as scalar_sqrt
from numpy import \
    array, zeros, outer, inner, transpose, dot, \
    fabs, identity, tensordot, \
    float64, \
    sqrt as arr_sqrt, copy
from scipy.linalg import \
    eigh, inv
from traits.api import \
    Bool, Callable, Enum, \
    Int, Trait, on_trait_change, \
    Dict, Property, cached_property

from .matsXD_cmdm_polar_discr import \
    PolarDiscr
import numpy as np


# @todo parameterize for 2D and 2D5 - does not make sense for 3D
#---------------------------------------------------------------------------
# Material time-step-evaluator for Microplane-Damage-Model
#---------------------------------------------------------------------------
class MATSXDMicroplaneDamage(PolarDiscr):

    '''
    Microplane Damage Model.
    '''

    #-------------------------------------------------------------------------
    # Classification traits (supplied by the dimensional subclasses)
    #-------------------------------------------------------------------------

    # specification of the model dimension (2D, 3D)
    n_dim = Int

    # specification of number of engineering strain and stress components
    n_eng = Int

    #-------------------------------------------------------------------------
    # Configuration parameters
    #-------------------------------------------------------------------------

    model_version = Enum("compliance", "stiffness")

    symmetrization = Enum("product-type", "sum-type")

    regularization = Bool(False,
                          desc='Flag to use the element length projection'
                          ' in the direction of principle strains',
                          enter_set=True,
                          auto_set=False)

    elastic_debug = Bool(False,
                         desc='Switch to elastic behavior - used for debugging',
                         auto_set=False)

    double_constraint = Bool(False,
                             desc='Use double constraint to evaluate microplane elastic and fracture energy (Option effects only the response tracers)',
                             auto_set=False)

    def get_state_array_size(self):
        # In the state array the largest equivalent microplane
        # strains reached in the loading history are saved
        return self.n_mp

    D4_e = Property

    def _get_D4_e(self):
        '''
        Return the elasticity tensor
        '''
        return self.elasticity_tensors[0]

    C4_e = Property

    def _get_C4_e(self):
        '''
        Return the elastic compliance tensor
        '''
        return self.elasticity_tensors[1]

    # -----------------------------------------------------------------------------------------------------
    # Get the 3x3-elasticity and compliance matrix (used only for elastic debug)
    # -----------------------------------------------------------------------------------------------------
    D2_e = Property(depends_on='stress_state, E, nu')

    @cached_property
    def _get_D2_e(self):
        return self.elasticity_tensors[2]

    identity_tns = Property

    @cached_property
    def _get_identity_tns(self):
        '''
        Get the identity matrix (used only in formula for sum-type symmetrization)
        ''' 
        return identity(self.n_dim)

    #-------------------------------------------------------------------------
    # MICROPLANE-DISCRETIZATION RELATED METHOD
    #-------------------------------------------------------------------------
    # get the dyadic product of the microplane normals
    # the array of microplane normals is implemented
    # in the dimension-specific subclasses.
    #
    _MPNN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPNN(self):
        # dyadic product of the microplane normals
        return array([outer(mpn, mpn) for mpn in self._MPN])

    #-------------------------------------------------------------------------
    # Prepare evaluation - get the 4th order damage tensor beta4
    # (or damage effect tensor M4)
    #-------------------------------------------------------------------------
    def _get_e_vct_arr(self, eps_eng):
        '''
        Projects the strain tensor onto the microplanes and returns a list of
        microplane strain vectors. Method is used both by stiffness and
        compliance version to derive the list 'phi_arr' or 'psi_arr'!
        In case of the compliance version the kinematic constraint is not
        assumed in the derivation of the formula for the damage compliance
        tensor 'C_mdm', e.g. the construction of the damage effect tensor 'M4'.
        '''
        # Switch from engineering notation to tensor notation for the apparent
        # strains
        eps_mtx = self.map_eps_eng_to_mtx(eps_eng)
        # Projection of apparent strain onto the individual microplanes
        # slower: e_vct_arr = array( [ dot( eps_mtx, mpn ) for mpn in self._MPN ] )
        # slower: e_vct_arr = transpose( dot( eps_mtx, transpose(self._MPN) ))
        # due to the symmetry of the strain tensor eps_mtx = transpose(eps_mtx)
        # and so this is equal
        e_vct_arr = dot(self._MPN, eps_mtx)
        return e_vct_arr

    def _get_e_equiv_arr(self, e_vct_arr):
        '''
        Returns a list of the microplane equivalent strains
        based on the list of microplane strain vectors
        '''
        # magnitude of the normal strain vector for each microplane
        # @todo: faster numpy functionality possible?
        e_N_arr = array([dot(e_vct, mpn)
                         for e_vct, mpn in zip(e_vct_arr, self._MPN)])
        # positive part of the normal strain magnitude for each microplane
        e_N_pos_arr = (fabs(e_N_arr) + e_N_arr) / 2
        # normal strain vector for each microplane
        # @todo: faster numpy functionality possible?
        e_N_vct_arr = array([self._MPN[i, :] * e_N_arr[i]
                             for i in range(0, self.n_mp)])
        # tangent strain ratio
        c_T = self.c_T
        # tangential strain vector for each microplane
        e_T_vct_arr = e_vct_arr - e_N_vct_arr
        # squared tangential strain vector for each microplane
        e_TT_arr = array([inner(e_T_vct, e_T_vct) for e_T_vct in e_T_vct_arr])
        # equivalent strain for each microplane
        e_equiv_arr = arr_sqrt(e_N_pos_arr * e_N_pos_arr + c_T * e_TT_arr)
        return e_equiv_arr

    def _get_e_max(self, e_equiv_arr, e_max_arr):
        '''
        Compares the equivalent microplane strain of a single microplane with the
        maximum strain reached in the loading history for the entire array
        '''
        bool_e_max = e_equiv_arr >= e_max_arr

        # [rch] fixed a bug here - this call was modifying the state array
        # at any invocation.
        #
        # The new value must be created, otherwise side-effect could occur
        # by writing into a state array.
        #
        new_e_max_arr = copy(e_max_arr)
        new_e_max_arr[bool_e_max] = e_equiv_arr[bool_e_max]
        return new_e_max_arr

    def _get_state_variables(self, sctx, eps_app_eng):
        '''
        Compares the list of current equivalent microplane strains with
        the values in the state array and returns the higher values
        '''
        e_vct_arr = self._get_e_vct_arr(eps_app_eng)
        e_equiv_arr = self._get_e_equiv_arr(e_vct_arr)
        e_max_arr_old = sctx.mats_state_array
        e_max_arr_new = self._get_e_max(e_equiv_arr, e_max_arr_old)
        return e_max_arr_new

    def _get_phi_arr(self, sctx, eps_app_eng):
        '''
        Returns a list of the integrity factors for all microplanes.
        '''
        e_max_arr = self._get_state_variables(sctx, eps_app_eng)

        # @todo: is this a possible position to add in the if-case evaluation
        # for compression in case of a z-direction "damage broadcasting"?!
        #
        # get the sign of the normal component of each microplane
        # i.e. check if the microplane strain is negative (compression)
        # if yes ignore the state variable and explicitly set the phi value to 1.0
        # which corresponds to the initial (undamaged) value of e_max = 0.
        # magnitude of the normal strain vector for each microplane
        # @todo: faster numpy functionality possible? Note that the calculation of 'e_N_arr' is calculated twice!
        #
#        e_vct_arr = self._get_e_vct_arr(eps_app_eng)
#        e_N_arr = array([ dot(e_vct, mpn) for e_vct, mpn in zip(e_vct_arr, self._MPN) ])
#        bool_arr = e_N_arr < 0.
#        e_max_arr[bool_arr] = 0.

        return self.get_phi_arr(sctx, e_max_arr)

    def _get_phi_mtx(self, sctx, eps_app_eng):
        '''
        Returns the 2nd order damage tensor 'phi_mtx'
        '''
        # scalar integrity factor for each microplane
        phi_arr = self._get_phi_arr(sctx, eps_app_eng)
        # integration terms for each microplanes
        # @todo: faster numpy functionality possible?
        phi_mtx_arr = array([phi_arr[i] * self._MPNN[i, :, :] * self._MPW[i]
                             for i in range(0, self.n_mp)])
        # sum of contributions from all microplanes
        # sum over the first dimension (over the microplanes)
        phi_mtx = phi_mtx_arr.sum(0)
        return phi_mtx

    def _get_psi_mtx(self, sctx, eps_app_eng):
        '''
        Returns the 2nd order damage effect tensor 'psi_mtx'
        '''
        # scalar integrity factor for each microplane
        phi_arr = self._get_phi_arr(sctx, eps_app_eng)
        # integration terms for each microplanes

        psi_mtx_arr = array([1. / phi_arr[i] * self._MPNN[i, :, :] * self._MPW[i]
                             for i in range(0, self.n_mp)])

        # sum of contributions from all microplanes
        # sum over the first dimension (over the microplanes)
        psi_mtx = psi_mtx_arr.sum(0)
        return psi_mtx

    def _get_beta_tns_product_type(self, phi_mtx):
        '''
        Returns the 4th order damage tensor 'beta4' using product-type symmetrization
        (cf. [Baz97], Eq.(87))
        '''
        n_dim = self.n_dim
        # Get the direction of the principle damage coordinates (pdc):
        phi_eig_value, phi_eig_mtx = eigh(phi_mtx)
        phi_eig_value_real = array([pe.real for pe in phi_eig_value])
        # transform phi_mtx to PDC:
        # (assure that besides the diagonal the entries are exactly zero)
        phi_pdc_mtx = zeros((n_dim, n_dim), dtype=float)
        for i in range(n_dim):
            phi_pdc_mtx[i, i] = phi_eig_value_real[i]
        # w_mtx = tensorial square root of the second order damage tensor:
        w_pdc_mtx = arr_sqrt(phi_pdc_mtx)
#        print "w_pdc_mtx", w_pdc_mtx
        # transform the matrix w back to x-y-coordinates:
        w_mtx = dot(dot(phi_eig_mtx, w_pdc_mtx), transpose(phi_eig_mtx))
        # beta_ijkl = w_ik * w_jl (cf. [Baz 97])
        # exploiting numpy-functionality (faster).
        # Method 'outer' returns beta_ijkl = w_ij * w_kl,
        # therefore the axis j and k need to be swapped
        beta4_ = outer(w_mtx, w_mtx).reshape(n_dim, n_dim, n_dim, n_dim)
        beta4 = beta4_.swapaxes(1, 2)
        return beta4

    def _get_beta_tns_sum_type(self, phi_mtx):
        '''
        Returns the 4th order damage tensor 'beta4' using sum-type symmetrization
        (cf. [Jir99], Eq.(21))
        '''
        n_dim = self.n_dim
        delta = self.identity_tns

        # The following line correspond to the tensorial expression:
        #
        #        beta4 = zeros((n_dim,n_dim,n_dim,n_dim),dtype=float)
        #        for i in range(0,n_dim):
        #            for j in range(0,n_dim):
        #                for k in range(0,n_dim):
        #                    for l in range(0,n_dim):
        #                        beta4[i,j,k,l] = 0.25 * ( phi_mtx[i,k] * delta[j,l] + phi_mtx[i,l] * delta[j,k] +\
        #                                                  phi_mtx[j,k] * delta[i,l] + phi_mtx[j,l] * delta[i,k] )
        #

        # use numpy functionality to evaluate [Jir99], Eq.(21)
        beta_ijkl = outer(phi_mtx, delta).reshape(n_dim, n_dim, n_dim, n_dim)
        beta_ikjl = beta_ijkl.swapaxes(1, 2)
        beta_iljk = beta_ikjl.swapaxes(2, 3)
        beta_jlik = beta_iljk.swapaxes(0, 1)
        beta_jkil = beta_jlik.swapaxes(2, 3)
        beta4 = 0.25 * (beta_ikjl + beta_iljk + beta_jkil + beta_jlik)

        return beta4

    def _get_M_tns_product_type(self, psi_mtx):
        '''
        Returns the 4th order damage effect tensor 'M4' using product-type symmetrization
        '''
        n_dim = self.n_dim
        # Get the direction orthogonal to the principle damage coordinates (pdc):
        # @todo: is this direction orthogonal? Which one do we want?
        psi_eig_value, psi_eig_mtx = eigh(psi_mtx)
        psi_eig_value_real = array([pe.real for pe in psi_eig_value])
        # transform phi_mtx to PDC:
        # (assure that besides the diagonal the entries are exactly zero)
        psi_pdc_mtx = zeros((n_dim, n_dim), dtype=float)
        for i in range(n_dim):
            psi_pdc_mtx[i, i] = psi_eig_value_real[i]
        # second order damage effect tensor:
        w_hat_pdc_mtx = arr_sqrt(psi_pdc_mtx)
#        print "w_hat_pdc_mtx", w_hat_pdc_mtx
        # transform the matrix w back to x-y-coordinates:
        w_hat_mtx = dot(
            dot(psi_eig_mtx, w_hat_pdc_mtx), transpose(psi_eig_mtx))
#        print "w_hat_mtx", w_hat_mtx
        # M_ijkl = w_hat_ik * w_hat_lj (cf. Eq.(5.62) Script Prag Jirasek (2007))
        #        = w_hat_ik * w_hat_jl (w is a symmetric tensor)
        # Exploiting numpy-functionality using the
        # method 'outer' (returns M_ijkl = w_hat_ij * w_hat_kl),
        # therefore the axis j and k need to be swapped
        M4_ = outer(w_hat_mtx, w_hat_mtx).reshape(n_dim, n_dim, n_dim, n_dim)
        M4 = M4_.swapaxes(1, 2)
        return M4

    def _get_M_tns_sum_type(self, psi_mtx):
        '''
        Returns the 4th order damage effect tensor 'M4' using sum-type symmetrization
        '''
        n_dim = self.n_dim
        delta = self.identity_tns

        # The line below corresponds to the tensorial expression
        #
        # (cf. [Jir99], Eq.(30))
        #        M4 = zeros((n_dim,n_dim,n_dim,n_dim),dtype=float)
        #        for i in range(0,n_dim):
        #            for j in range(0,n_dim):
        #                for k in range(0,n_dim):
        #                    for l in range(0,n_dim):
        #                        M4[i,j,k,l] = 0.25 * ( psi_mtx[i,k] * delta[j,l] + psi_mtx[i,l] * delta[j,k] +\
        # psi_mtx[j,k] * delta[i,l] + psi_mtx[j,l] * delta[i,k] )

        # use numpy functionality to evaluate [Jir99], Eq.(21)
        M_ijkl = outer(psi_mtx, delta).reshape(n_dim, n_dim, n_dim, n_dim)
        M_ikjl = M_ijkl.swapaxes(1, 2)
        M_iljk = M_ikjl.swapaxes(2, 3)
        M_jlik = M_iljk.swapaxes(0, 1)
        M_jkil = M_jlik.swapaxes(2, 3)
        M4 = 0.25 * (M_ikjl + M_iljk + M_jkil + M_jlik)

        return M4

    #-------------------------------------------------------------------------
    # Dynamic configuration of the damage tensor evaluation
    #-------------------------------------------------------------------------
    # the traits _get_beta_tns and _get_M_tns are are reset
    # depending on the selected type of symmetrization.

    _get_beta_tns = Callable(transient=True)

    def __get_beta_tns_default(self):
        return self._get_damage_eval_methods()[0]

    _get_M_tns = Callable(transient=True)

    def __get_M_tns_default(self):
        return self._get_damage_eval_methods()[1]

    @on_trait_change('symmetrization')
    def _reset_damage_eval_methods(self):
        self._get_beta_tns, self._get_M_tns = self._get_damage_eval_methods()

    def _get_damage_eval_methods(self):
        if self.symmetrization == 'product-type':
            return (self._get_beta_tns_product_type, self._get_M_tns_product_type)
        elif self.symmetrization == 'sum-type':
            return (self._get_beta_tns_sum_type, self._get_M_tns_sum_type)
        else:
            raise ValueError('Bad symmetrization tag')

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------
    def get_corr_pred(self, sctx, eps_app_eng, d_eps, tn, tn1, eps_avg=None):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''

        # -----------------------------------------------------------------------------------------------
        # check if average strain is to be used for damage evaluation
        # -----------------------------------------------------------------------------------------------
        if eps_avg != None:
            pass
        else:
            eps_avg = eps_app_eng

        # -----------------------------------------------------------------------------------------------
        # for debugging purposes only: if elastic_debug is switched on, linear elastic material is used
        # -----------------------------------------------------------------------------------------------
        if self.elastic_debug:
            # NOTE: This must be copied otherwise self.D2_e gets modified when
            # essential boundary conditions are inserted
            D2_e = copy(self.D2_e)
            sig_eng = tensordot(D2_e, eps_app_eng, [[1], [0]])
            return sig_eng, D2_e

        # -----------------------------------------------------------------------------------------------
        # update state variables
        # -----------------------------------------------------------------------------------------------
        if sctx.update_state_on:
            eps_n = eps_avg - d_eps
            e_max = self._get_state_variables(sctx, eps_n)
            sctx.mats_state_array[:] = e_max

        #----------------------------------------------------------------------
        # if the regularization using the crack-band concept is on calculate the
        # effective element length in the direction of principle strains
        #----------------------------------------------------------------------
        if self.regularization:
            h = self.get_regularizing_length(sctx, eps_app_eng)
            self.phi_fn.h = h

        #----------------------------------------------------------------------
        # stiffness version:
        #----------------------------------------------------------------------
        if self.model_version == 'stiffness':

            #------------------------------------------------------------------
            # Damage tensor (2th order):
            #------------------------------------------------------------------

            phi_mtx = self._get_phi_mtx(sctx, eps_avg)

            #------------------------------------------------------------------
            # Damage tensor (4th order) using product- or sum-type symmetrization:
            #------------------------------------------------------------------
            beta4 = self._get_beta_tns(phi_mtx)

            #------------------------------------------------------------------
            # Damaged stiffness tensor calculated based on the damage tensor beta4:
            #------------------------------------------------------------------
            # (cf. [Jir99] Eq.(7): C = beta * D_e * beta^T),
            # minor symmetry is tacitly assumed ! (i.e. beta_ijkl = beta_jilk)
            D4_mdm = tensordot(
                beta4, tensordot(self.D4_e, beta4, [[2, 3], [2, 3]]), [[2, 3], [0, 1]])

            #------------------------------------------------------------------
            # Reduction of the fourth order tensor to a matrix assuming minor and major symmetry:
            #------------------------------------------------------------------
            D2_mdm = self.map_tns4_to_tns2(D4_mdm)

        #----------------------------------------------------------------------
        # compliance version:
        #----------------------------------------------------------------------
        elif self.model_version == 'compliance':

            #------------------------------------------------------------------
            # Damage effect tensor (2th order):
            #------------------------------------------------------------------

            psi_mtx = self._get_psi_mtx(sctx, eps_avg)

            #------------------------------------------------------------------
            # Damage effect tensor (4th order) using product- or sum-type-symmetrization:
            #------------------------------------------------------------------
            M4 = self._get_M_tns(psi_mtx)

            #------------------------------------------------------------------
            # Damaged compliance tensor calculated based on the damage effect tensor M4:
            #------------------------------------------------------------------
            # (cf. [Jir99] Eq.(8): C = M^T * C_e * M,
            # minor symmetry is tacitly assumed ! (i.e. M_ijkl = M_jilk)
            C4_mdm = tensordot(
                M4, tensordot(self.C4_e, M4, [[2, 3], [0, 1]]), [[0, 1], [0, 1]])

            #------------------------------------------------------------------
            # Reduction of the fourth order tensor to a matrix assuming minor and major symmetry:
            #------------------------------------------------------------------
            C2_mdm = self.map_tns4_to_tns2(C4_mdm)
            D2_mdm = inv(self.compliance_mapping(C2_mdm))

        #----------------------------------------------------------------------
        # Return stresses (corrector) and damaged secant stiffness matrix (predictor)
        #----------------------------------------------------------------------

        sig_eng = tensordot(D2_mdm, eps_app_eng, [[1], [0]])
        return sig_eng, D2_mdm

    #-------------------------------------------------------------------------
    # Control variables and update state method
    #-------------------------------------------------------------------------
    def new_cntl_var(self):
        return zeros(self.n_eng, float64)

    def new_resp_var(self):
        return zeros(self.n_eng, float64)

    def update_state(self, sctx, eps_app_eng):
        '''
        Update state method is called upon an accepted time-step.
        Here just set the flag on to make the update afterwards in the method itself.
        '''
        print('in update-state')
        # self.update_state_on = True

    #-------------------------------------------------------------------------
    # Response trace evaluators
    #-------------------------------------------------------------------------
    def get_eps_app_v3d(self, sctx, eps_app_eng):
        return eps_app_eng

    def get_sig_app_v3d(self, sctx, eps_app_eng, *args, **kw):
        # @TODO
        # the stress calculation is performed twice - it might be
        # cached. But not in the spatial integration scheme.
        sig_app_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return sig_app_eng

    def get_eps_app(self, sctx, eps_app_eng):
        return self.map_eps_eng_to_mtx(eps_app_eng)

    def get_sig_app(self, sctx, eps_app_eng, *args, **kw):
        # @TODO
        # the stress calculation is performed twice - it might be
        # cached. But not in the spatial integration scheme.
        sig_app_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        # return sig_app_eng
        return self.map_sig_eng_to_mtx(sig_app_eng)

    def get_microplane_integrity(self, sctx, eps_app_eng):
        phi_arr = self._get_phi_arr(sctx, eps_app_eng)
        return phi_arr

    def get_sig_norm(self, sctx, eps_app_eng):
        sig_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        return array([scalar_sqrt(sig_eng[0] ** 2 + sig_eng[1] ** 2)])

    def get_phi_mtx(self, sctx, eps_app_eng):
        return self._get_phi_mtx(sctx, eps_app_eng)

    def get_phi_pdc(self, sctx, eps_app_eng):
        phi_mtx = self._get_phi_mtx(sctx, eps_app_eng)
        # Get the direction of the principle damage coordinates (pdc):
        phi_eig_value, phi_eig_mtx = eigh(phi_mtx)
        phi_eig_value_real = array([pe.real for pe in phi_eig_value])
        # return the minimum value of the eigenvalues of the integrity tensor
        # (= values of integrity in the principle direction)
        #
        min_phi = np.min(phi_eig_value_real)
        max_omega = (1.0 - min_phi)
        return array([max_omega])

    def get_phi_pdc2(self, sctx, eps_app_eng):
        phi_mtx = self._get_phi_mtx(sctx, eps_app_eng)
        # Get the direction of the principle damage coordinates (pdc):
        phi_eig_value, phi_eig_mtx = eigh(phi_mtx)
        phi_eig_value_real = array([pe.real for pe in phi_eig_value])
        # return the minimum value of the eigenvalues of the integrity tensor
        # (= values of integrity in the principle direction)
        #
        min_phi = np.min(phi_eig_value_real)
        max_omega = (1.0 - min_phi ** 2)
        return array([max_omega])

    # ------------------------------------------
    # SUBSIDARY METHODS used only for the response tracer:
    # ------------------------------------------
    # Projection: Methods used for projection:

    #  '_get_e_vct_arr'  -  has been defined above (see 'get_corr_pred')
    #   (also used as subsidary method for '_get_e_s_vct_arr'.)

    def _get_s_vct_arr(self, sig_eng):
        '''
        Projects the stress tensor onto the microplanes and
        returns a list of microplane strain vectors.
        (Subsidary method for '_get_e_s_vct_arr'.)
        '''
        # Switch from engineering notation to tensor notation for the apparent
        # strains
        sig_mtx = self.map_sig_eng_to_mtx(sig_eng)
        # Projection of apparent strain onto the individual microplanes
        # slower: s_vct_arr = array( [ dot( sig_mtx, mpn ) for mpn in self._MPN
        # ] )
        s_vct_arr = dot(self._MPN, sig_mtx)
        return s_vct_arr

    # Equiv: Equivalent microplane parts:

    #  '_get_e_equiv_arr'  -  has been defined above (see 'get_corr_pred')

    def _get_s_equiv_arr(self, s_vct_arr):
        '''
        Get microplane equivalent stress.
        (Subsidary method for '_get_e_s_equiv_arr'.)
        '''
        # The same method is used to calculate the equivalent stresses
        # and the equivalent strains
        s_equiv_arr = self._get_e_equiv_arr(s_vct_arr)
        return s_equiv_arr

    # N: normal microplane parts

    def _get_e_N_arr(self, e_vct_arr):
        '''
        Returns a list of the microplane normal strains (scalar)
        based on the list of microplane strain vectors
        (Subsidary method for '_get_e_s_N_arr'.) '''
        # magnitude of the normal strain vector for each microplane
        e_N_arr = array([dot(e_vct, mpn)
                         for e_vct, mpn in zip(e_vct_arr, self._MPN)])
        return e_N_arr
        # @todo: check if the direct calculation of e_N using MPNN makes sense.
        #        Note that e_T needs the calculation of e_N_vct !
        #        s_N_arr = array( [ s_vct_arr[i] * self._MPNN[i,:,:] for i in range(0,self.n_mp) ] )
        #        return s_N_arr[0,:].flatten()

    def _get_s_N_arr(self, s_vct_arr):
        # the same method is used for the calculation of the
        # normal parts of the microplane strain and stress vector
        s_N_arr = self._get_e_N_arr(s_vct_arr)
        return s_N_arr

    # T: tangential microplane parts

    def _get_e_T_arr(self, e_vct_arr):
        '''
        Returns a list of the microplane shear strains (scalar)
        based on the list of microplane strain vectors
        (Subsidary method for '_get_e_s_T_arr'.)
        '''
        # magnitude of the normal strain vector for each microplane
        e_N_arr = self._get_e_N_arr(e_vct_arr)
        # normal strain vector for each microplane
        e_N_vct_arr = array([self._MPN[i, :] * e_N_arr[i]
                             for i in range(0, self.n_mp)])
        # tangential strain vector for each microplane
        e_T_vct_arr = e_vct_arr - e_N_vct_arr
        # squared tangential strain vector for each microplane
        e_TT_arr = array([inner(e_T_vct, e_T_vct) for e_T_vct in e_T_vct_arr])
        # equivalent strain for each microplane
        e_T_arr = arr_sqrt(e_TT_arr)
        return e_T_arr

    def _get_s_T_arr(self, s_vct_arr):
        # the same method is used for the calculation of the
        # tangential parts of the microplane strain and stress vector
        s_T_arr = self._get_e_T_arr(s_vct_arr)
        return s_T_arr

    # -------------------------------------------------------------
    # Get the microplane strain and stresses either based on a
    # consistently derived model-version or the double constraint
    # -------------------------------------------------------------
    def _get_e_s_vct_arr(self, sctx, eps_app_eng):
        '''
        Returns two arrays containing the microplane strain and stress vectors
        either assuming a double constraint or consistently derived based on the specified model version
        '''
        #----------------------------------------------------------------------
        # DOUBLE CONSTRAINT
        #----------------------------------------------------------------------
        if self.double_constraint:
            '''
            Return a pair of microplane stress and strain vectors based on the
            double constrain, i.e kinematic AND static constraint !
            The connection between the apparent strain and stress tensor is established
            with D_mdm based on the chosen model version (e.g. stiffness or compliance)
            '''
            # microplane strain vectors obtained by projection (kinematic
            # constraint):
            e_app_vct_arr = self._get_e_vct_arr(eps_app_eng)
            # get the corresponding macroscopic stresses
            sig_app_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
            # microplane stress vectors obtained by projection (static
            # constraint)
            s_app_vct_arr = self._get_s_vct_arr(sig_app_eng)
            return e_app_vct_arr, s_app_vct_arr

        #----------------------------------------------------------------------
        # CONSISTENTLY DERIVED pair of microplane strain and stress vectors
        #----------------------------------------------------------------------
        '''
        Returns two arrays containing the microplane strain and stress vectors
        consistently derived based on the specified model version, i.e. either 'stiffness' or 'compliance'
        '''
        #-------------------
        # stiffness version:
        #-------------------
        if self.model_version == 'stiffness':
            # microplane equivalent strains obtained by projection (kinematic
            # constraint)
            e_app_vct_arr = self._get_e_vct_arr(eps_app_eng)

            # microplane equivalent stresses calculated based on corresponding 'beta' and 'phi_mtx'
            # 2nd order damage tensor:
            phi_mtx = self._get_phi_mtx(sctx, eps_app_eng)
            # 4th order damage tensor:
            if self.symmetrization == 'product-type':
                beta4 = self._get_beta_tns_product_type(phi_mtx)
            elif self.symmetrization == 'sum-type':
                beta4 = self._get_beta_tns_sum_type(phi_mtx)

            # apparent strain tensor:
            eps_app_mtx = self.map_eps_eng_to_mtx(eps_app_eng)
            # effective strain tensor:
            eps_eff_mtx = tensordot(beta4, eps_app_mtx, [[0, 1], [0, 1]])
            # effective stress tensor:
            sig_eff_mtx = tensordot(self.D4_e, eps_eff_mtx, [[2, 3], [0, 1]])
            # effective microplane stresses obtained by projection (static
            # constraint)
            s_eff_vct_arr = array([dot(sig_eff_mtx, mpn) for mpn in self._MPN])
            # microplane scalar damage variable (integrity):
            phi_arr = self._get_phi_arr(sctx, eps_app_eng)
            # apparent microplane stresses
            s_app_vct_arr = array(
                [dot(phi, s_eff_vct) for phi, s_eff_vct in zip(phi_arr, s_eff_vct_arr)])

        #--------------------
        # compliance version:
        #--------------------
        if self.model_version == 'compliance':
            # get the corresponding macroscopic stresses
            sig_app_eng, D_mtx = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)

            # microplane equivalent stress obtained by projection (static
            # constraint)
            s_app_vct_arr = self._get_s_vct_arr(sig_app_eng)

            # microplane equivalent strains calculated based on corresponding 'M' and 'psi_mtx'
            # 2nd order damage effect tensor:
            psi_mtx = self._get_psi_mtx(sctx, eps_app_eng)
            # 4th order damage effect tensor:
            if self.symmetrization == 'product-type':
                M4 = self._get_M_tns_product_type(psi_mtx)
            elif self.symmetrization == 'sum-type':
                M4 = self._get_M_tns_sum_type(psi_mtx)

            # apparent stress tensor:
            sig_app_mtx = self.map_sig_eng_to_mtx(sig_app_eng)
            # effective stress tensor:
            sig_eff_mtx = tensordot(M4, sig_app_mtx, [[2, 3], [0, 1]])
            # effective strain tensor:
            eps_eff_mtx = tensordot(self.C4_e, sig_eff_mtx, [[2, 3], [0, 1]])
            # effective microplane strains obtained by projection (kinematic
            # constraint)
            e_eff_vct_arr = array([dot(eps_eff_mtx, mpn) for mpn in self._MPN])
            # microplane scalar damage variable (integrity):
            phi_arr = self._get_phi_arr(sctx, eps_app_eng)
            psi_arr = 1. / phi_arr
            # apparent microplane strains
            e_app_vct_arr = array(
                [dot(psi, e_eff_vct) for psi, e_eff_vct in zip(psi_arr, e_eff_vct_arr)])

        return e_app_vct_arr, s_app_vct_arr

    def _get_psi_arr_cv(self, sctx, e_max_arr):
        '''
        Returns a list of the compliance integrity factors for all microplanes, i.e. psi_i(e_max) = 1/phi_i(e_max)
        '''
        return 1. / self.get_phi_arr(sctx, e_max_arr)

    def _get_psi_mtx_cv(self, sctx, e_max_arr):
        '''
        Returns the 2nd order damage effect tensor 'psi_mtx'
        '''
        # scalar integrity factor for each microplane
        psi_arr = self._get_psi_arr_cv(sctx, e_max_arr)
        # integration terms for each microplanes
        psi_mtx_arr = array([psi_arr[i] * self._MPNN[i, :, :] * self._MPW[i]
                             for i in range(0, self.n_mp)])
        # sum of contributions from all microplanes
        # sum over the first dimension (over the microplanes)
        psi_mtx = psi_mtx_arr.sum(0)
        return psi_mtx

    #-------------------------------------------------------------------------
    #-------------------------------------------------------------------------
    # @todo: is an implicit formulation of CMDM possible? First outline see below
    #-------------------------------------------------------------------------
    #
    def get_lack_of_fit_psi_arr(self, sctx, e_max_arr_new, eps_app_eng):
        # state variables of last converged step at time t_i
        e_max_arr_old = sctx.mats_state_array

        # compare current values with state variables take max value in history
        e_max_arr = max(e_max_arr_old, e_max_arr_new)

        # get the corresponding macroscopic stresses for D_mtx at time t_i+1
        # for strain
        sig_app_eng, D_mtx_cv = self.get_corr_pred(
            sctx, e_max_arr, eps_app_eng, 0, 0, 0)
        psi_arr = self._get_psi_arr_cv(sctx, e_max_arr)
        psi_mtx_cv = self._get_psi_mtx(sctx, e_max_arr)
        if self.symmetrization == 'product-type':
            M4 = self._get_M_tns_product_type(psi_mtx_cv)
        elif self.symmetrization == 'sum-type':
            M4 = self._get_M_tns_sum_type(psi_mtx_cv)
        # apparent stress tensor:
        sig_app_mtx = self.map_sig_eng_to_mtx(sig_app_eng)
        # effective stress tensor:
        sig_eff_mtx = tensordot(M4, sig_app_mtx, [[2, 3], [0, 1]])
        # effective strain tensor:
        eps_eff_mtx = tensordot(self.C4_e, sig_eff_mtx, [[2, 3], [0, 1]])
        # effective microplane strains obtained by projection (kinematic
        # constraint)
        e_eff_vct_arr = array([dot(eps_eff_mtx, mpn) for mpn in self._MPN])
        # microplane scalar damage variable (integrity):
        # apparent microplane strains
        e_app_vct_arr = array([dot(psi, e_eff_vct)
                               for psi, e_eff_vct in zip(psi_arr, e_eff_vct_arr)])
        # equivalent strain for each microplane
        e_equiv_arr = self._get_e_equiv_arr(e_app_vct_arr)
        # lack of fit in the state variables
        psi_arr_trial = self._get_psi_arr_cv(sctx, e_equiv_arr)
        lof = psi_arr_trial - psi_arr
        return lof

    # -------------------------------------------------------------
    # Get the values of the maximum microplane strains (state variables) iteratively
    # calculated within each iteration step of the global time loop algorithm fullfilling
    # also in the trial steps the consistently derived damage tensor based on the compliance
    # version (implicit formulation)
    # -------------------------------------------------------------
    def get_corr_pred_cv(self, sctx, eps_app_eng, d_eps, tn, tn1, eps_avg=None):
        '''
        Returns two arrays containing the microplane strain and stress vectors
        consistently derived based on the specified model version, i.e. either 'stiffness' or 'compliance'
        '''
        # for compliance version only 
        #
        if self.model_version != 'compliance':
            raise ValueError('only valid for compliance version')

        raise ImportError('If this is needed include the import of the scipy package')
        #e_max_arr_new = brentq(e_max_arr_new, self.get_lack_of_fit_psi_arr(
        #    self, sctx, e_msax_arr_new, eps_app_eng))

        #self.update_state_variables(e_max_arr_new)

        # get the corresponding macroscopic stresses for D_mtx at time t_i+1
        # for strain
        #sig_app_eng, D_mtx_cv = self.get_corr_pred(
        #    sctx, e_max_arr_new, eps_app_eng, 0, 0, 0)

    #-------------------------------------------------------------------------
    #-------------------------------------------------------------------------

    # Equiv:
    def _get_e_s_equiv_arr(self, sctx, eps_app_eng):
        '''
        Returns two arrays containing the corresponding equivalent
        microplane strains and stresses.
        '''
        e_app_vct_arr, s_app_vct_arr = self._get_e_s_vct_arr(sctx, eps_app_eng)

        # equivalent strain for each microplane
        e_equiv_arr = self._get_e_equiv_arr(e_app_vct_arr)
        # equivalent strain for each microplane
        s_equiv_arr = self._get_s_equiv_arr(s_app_vct_arr)
        return e_equiv_arr, s_equiv_arr

    def get_e_equiv_arr(self, sctx, eps_app_eng):
        '''
        Return a list of equivalent microplane strains.
        '''
        e_equiv_arr, s_equiv_arr = self._get_e_s_equiv_arr(sctx, eps_app_eng)
        return e_equiv_arr

    def get_s_equiv_arr(self, sctx, eps_app_eng):
        '''
        Return a list of equivalent microplane stresses.
        '''
        e_equiv_arr, s_equiv_arr = self._get_e_s_equiv_arr(sctx, eps_app_eng)
        return s_equiv_arr

    # N: normal microplane strain and stresses:
    def _get_e_s_N_arr(self, sctx, eps_app_eng):
        '''
        Returns two arrays containing the corresponding normal
        microplane strains and stress components.
        '''
        e_app_vct_arr, s_app_vct_arr = self._get_e_s_vct_arr(sctx, eps_app_eng)
        # normal strain for each microplane
        e_N_arr = self._get_e_N_arr(e_app_vct_arr)
        # normal strain for each microplane
        s_N_arr = self._get_s_N_arr(s_app_vct_arr)
        return e_N_arr, s_N_arr

    def get_e_N_arr(self, sctx, eps_app_eng):
        '''
        Return a list of normal microplane strains components.
        '''
        e_N_arr, s_N_arr = self._get_e_s_N_arr(sctx, eps_app_eng)
        return e_N_arr

    def get_s_N_arr(self, sctx, eps_app_eng):
        '''
        Return a list of normal microplane stresses components.
        '''
        e_N_arr, s_N_arr = self._get_e_s_N_arr(sctx, eps_app_eng)
        return s_N_arr

    # T: tangential microplane strain and stresses:
    def _get_e_s_T_arr(self, sctx, eps_app_eng):
        '''
        Returns two arrays containing the corresponding tangential
        microplane strains and stress components.
        '''
        e_app_vct_arr, s_app_vct_arr = self._get_e_s_vct_arr(sctx, eps_app_eng)
        # shear strain for each microplane
        e_T_arr = self._get_e_T_arr(e_app_vct_arr)
        # shear strain for each microplane
        s_T_arr = self._get_s_T_arr(s_app_vct_arr)
        return e_T_arr, s_T_arr

    def get_e_T_arr(self, sctx, eps_app_eng):
        '''
        Return a list of tangential microplane strains components.
        '''
        e_T_arr, s_T_arr = self._get_e_s_T_arr(sctx, eps_app_eng)
        return e_T_arr

    def get_s_T_arr(self, sctx, eps_app_eng):
        '''
        Return a list of tangential microplane stresses components.
        '''
        e_T_arr, s_T_arr = self._get_e_s_T_arr(sctx, eps_app_eng)
        return s_T_arr

# @todo: remove / old

# @todo: remove: this method and method 'get_integ' in 'polar_fn'
# old implementation: this assumes a decoupled reaction of all microplanes
    def get_fracture_energy_arr(self, sctx, e_max_arr):
        '''
        Get the microplane contributions to the fracture energy
        '''
        fracture_energy_arr = self.get_polar_fn_fracture_energy_arr(
            sctx, e_max_arr)
        return fracture_energy_arr

    def get_fracture_energy(self, sctx, eps_app_eng, *args, **kw):
        '''
        Get the macroscopic fracture energy as a weighted sum of all mircoplane contributions
        '''
        e_max_arr = self._get_state_variables(sctx, eps_app_eng)
        fracture_energy_arr = self.get_fracture_energy_arr(sctx, e_max_arr)
        fracture_energy = array([dot(self._MPW, fracture_energy_arr)], float)
        return fracture_energy

    def get_e_equiv_projection(self, sctx, eps_app_eng):
        '''
        Return a list of equivalent microplane strains as the projection of the strain tensor
        (kinematic constraint)
        '''
        e_vct_arr = self._get_e_vct_arr(eps_app_eng)
        e_equiv_arr = self._get_e_equiv_arr(e_vct_arr)
#        print 'e_equiv_arr:  ', e_equiv_arr
        return e_equiv_arr

    def get_max_omega_i2(self, sctx, eps_app_eng):
        '''
        Get maximum damage at all microplanes.
        '''
        min_phi = np.min(self._get_phi_arr(sctx, eps_app_eng))
        max_omega = 1. - min_phi ** 2
        if max_omega == 1.:
            print('max_omega_i2', max_omega)
#            print 'eps_app_eng', eps_app_eng
        return np.array([max_omega])

    def get_max_omega_i(self, sctx, eps_app_eng):
        '''
        Get maximum damage at all microplanes.
        '''
        min_phi = np.min(self._get_phi_arr(sctx, eps_app_eng))
        max_omega = 1. - min_phi
        if max_omega == 1.:
            print('max_omega_i', max_omega)
#            print 'eps_app_eng', eps_app_eng
        return np.array([max_omega])

    def get_omega_mtx(self, sctx, eps_app_eng, *args, **kw):
        '''
        Returns the 2nd order damage tensor 'phi_mtx'
        '''
        # scalar integrity factor for each microplane
        phi_arr = self._get_phi_arr(sctx, eps_app_eng)
        omega_arr = 1 - phi_arr ** 2
        # integration terms for each microplanes
        # @todo: faster numpy functionality possible?
        omega_mtx_arr = array([omega_arr[i] * self._MPNN[i, :, :] * self._MPW[i]
                               for i in range(0, self.n_mp)])
        # sum of contributions from all microplanes
        # sum over the first dimension (over the microplanes)
        omega_mtx = omega_mtx_arr.sum(0)

        return omega_mtx
# ##





from math import sin

from traits.api import \
    HasTraits, Float, provides, Interface


class ILevelSetFn(Interface):
    def level_set_fn(self, x, y):
        '''Level set function evaluation.
        '''
        raise NotImplementedError


@provides(ILevelSetFn)
class SinLSF(HasTraits):
    a = Float(1.5, enter_set=True, auto_set=False)
    b = Float(2.0, enter_set=True, auto_set=False)

    def level_set_fn(self, x, y):
        '''Level set function evaluation.
        '''
        return y - (sin(self.b * x) + self.a)


@provides(ILevelSetFn)
class PlaneLSF(HasTraits):
    a = Float(.5, enter_set=True, auto_set=False)
    b = Float(2.0, enter_set=True, auto_set=False)
    c = Float(-2.5, enter_set=True, auto_set=False)

    def level_set_fn(self, x, y):
        '''Level set function evaluation.
        '''
        return self.a * x + self.b * y + self.c


@provides(ILevelSetFn)
class ElipseLSF(HasTraits):
    a = Float(.5, enter_set=True, auto_set=False)
    b = Float(2.0, enter_set=True, auto_set=False)
    c = Float(-2.5, enter_set=True, auto_set=False)

    def level_set_fn(self, x, y):
        '''Level set function evaluation.
        '''
        return self.a * x * x + self.b * y * y - self.c




import copy
from ibvpy.sim.tmodel import TModel
from traits.api import \
    provides,  \
    Int, Str, Callable, \
    Tuple, Property, cached_property, \
    Dict, Interface
from ibvpy.view.ui import BMCSTreeNode

import numpy as np


class IMATSEval(Interface):
    '''Interface for material model classes
    '''

    state_var_shapes = Dict(Str, Tuple)

    var_dict = Dict(Str, Callable)

    explorer_rtrace_list = Property

    explorer_config = Property

#-------------------------------------------------------------------
# MATSEval - general implementation of the fe-numerical quadrature
#-------------------------------------------------------------------


@provides(IMATSEval)
class MATSEval(TModel, BMCSTreeNode):

    state_var_shapes = Dict(Str, Tuple, {})
    '''Shape of state variables
    '''

    #=========================================================================
    # Response variables
    #=========================================================================
    def get_eps(self, eps, tn1, **state):
        return eps

    def get_sig(self, eps, tn1, **state):
        state_copy = copy.deepcopy(state)
        sig, _ = self.get_corr_pred(
            eps, tn1, **state_copy
        )
        return sig

    var_dict = Property(Dict(Str, Callable))
    '''Dictionary of response variables
    '''
    @cached_property
    def _get_var_dict(self):
        return dict(eps=self.get_eps,
                    sig=self.get_sig)


class NotUsed:
    # Callable specifying spatial profile of an initial strain field
    # the parameter is X - global coordinates of the material point
    #
    initial_strain = Callable

    # Callable specifying spatial profile of an initial stress field
    # the parameter is X - global coordinates of the material point
    #
    initial_stress = Callable

    id_number = Int

    #-------------------------------------------------------------------------
    # Dimensionally dependent mappings between tensors
    #-------------------------------------------------------------------------
    # These are the handbook methods to be specialized in subclasses.

    # Mappings between tensorial and engineering variables
    #
    map_eps_eng_to_mtx = Callable(transient=True)
    map_sig_eng_to_mtx = Callable(transient=True)
    map_eps_mtx_to_eng = Callable(transient=True)
    map_sig_mtx_to_eng = Callable(transient=True)
    compliance_mapping = Callable(transient=True)
    map_tns4_to_tns2 = Callable(transient=True)

    def setup(self, sctx):
        pass

    explorer_rtrace_list = Property

    @cached_property
    def _get_explorer_rtrace_list(self):
        return []

    #-------------------------------------------------------------------------
    # Response trace evaluators
    #-------------------------------------------------------------------------
    def get_msig_pos(self, sctx, eps_app_eng, *args, **kw):
        '''
        get biggest positive principle stress
        @param sctx:
        @param eps_app_eng:
        '''
        sig_eng, _ = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        ms_vct = np.zeros(3)
        shape = sig_eng.shape[0]
        if shape == 3:
            s_mtx = self.map_sig_eng_to_mtx(sig_eng)
            m_sig, m_vct = np.linalg.eigh(s_mtx)

            # @todo: - this must be written in a more readable way
            #
            if m_sig[-1] > 0:
                # multiply biggest positive stress with its vector
                ms_vct[:2] = m_sig[-1] * m_vct[-1]
        elif shape == 6:
            s_mtx = self.map_sig_eng_to_mtx(sig_eng)
            m_sig = np.linalg.eigh(s_mtx)
            if m_sig[0][-1] > 0:
                # multiply biggest positive stress with its vector
                ms_vct = m_sig[0][-1] * m_sig[1][-1]
        return ms_vct

    def get_msig_pm(self, sctx, eps_app_eng, *args, **kw):
        sig_eng, _ = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        t_field = np.zeros(9)
        shape = sig_eng.shape[0]
        if shape == 3:
            s_mtx = self.map_sig_eng_to_mtx(sig_eng)
            m_sig = np.linalg.eigh(s_mtx)
            if m_sig[0][-1] > 0:
                t_field[0] = m_sig[0][-1]  # biggest positive stress
        elif shape == 6:
            s_mtx = self.map_sig_eng_to_mtx(sig_eng)
            m_sig = np.linalg.eigh(s_mtx)
            if m_sig[0][-1] > 0:
                t_field[0] = m_sig[0][-1]
        return t_field

    def get_max_principle_sig(self, sctx, eps_app_eng, *args, **kw):
        '''
        get biggest positive principle stress
        @param sctx:
        @param eps_app_eng:
        '''
        sig_eng, _ = self.get_corr_pred(sctx, eps_app_eng, 0, 0, 0)
        s_mtx = self.map_sig_eng_to_mtx(sig_eng)
        m_sig, _ = np.linalg.eigh(s_mtx)
        max_principle_sig = max(m_sig[:])
        # return max_principle_sig
        return np.array([max_principle_sig], dtype='float')

    def get_sig_app(self, sctx, eps_app_eng, *args, **kw):
        sig_eng, _ = self.get_corr_pred(
            sctx, eps_app_eng, 0, 0, 0, *args, **kw)
        s_tensor = np.zeros((3, 3))
        s_tensor[:self.n_dims, :self.n_dims] = self.map_sig_eng_to_mtx(sig_eng)
        return s_tensor

    def get_eps_app(self, sctx, eps_app_eng, *args, **kw):
        e_tensor = np.zeros((3, 3))
        e_tensor[:self.n_dims, :self.n_dims] = self.map_eps_eng_to_mtx(
            eps_app_eng)
        return e_tensor

    # This is only relevant for strain softening models
    #
    def get_regularizing_length(self, sctx, eps_app_eng, *args, **kw):

        X_mtx = sctx.X_reg

        # first principle strain unit vector
        #
        _, eigvec = np.linalg.eigh(self.map_eps_eng_to_mtx(eps_app_eng))

        # Get the eigenvector associated with maximum eigenvalue
        # it is located in the last column of the matrix
        # of eigenvectors eigvec
        #
        eps_one = eigvec[:, -1]
        # Project the coordinate vectors into the determined direction
        #
        proj = np.dot(X_mtx, eps_one)

        # Find the maximum distance between the projected coordinates
        #
        h = max(proj) - min(proj)
        return h

    def get_strain_energy(self, sctx, eps_app_eng, *args, **kw):
        sig_app = self.get_sig_app(sctx, eps_app_eng)
        eps_app = self.get_eps_app(sctx, eps_app_eng)
        energy = np.tensordot(sig_app, eps_app) * 0.5
        return energy

    # Declare and fill-in the explorer config
    # Each material model can define the default configuration
    # to present itself in the explorer.
    #
    explorer_config = Property(Dict)

    @cached_property
    def _get_explorer_config(self):
        from ibvpy.api import BCDof, TLine, RTDofGraph
        return {'bcond_list': [BCDof(var='u',
                                     dof=0, value=0.01,
                                     time_function=lambda t: t)],
                'rtrace_list': [RTDofGraph(name='strain - stress',
                                           var_x='eps_app', idx_x=0,
                                           var_y='sig_app', idx_y=0,
                                           record_on='update'),
                                RTDofGraph(name='strain - strain',
                                           var_x='eps_app', idx_x=0,
                                           var_y='eps_app', idx_y=1,
                                           record_on='update'),
                                RTDofGraph(name='stress - stress',
                                           var_x='sig_app', idx_x=0,
                                           var_y='sig_app', idx_y=1,
                                           record_on='update'),
                                RTDofGraph(name='Stress - Strain',
                                           var_x='F_int', idx_x=0,
                                           var_y='U_k', idx_y=0,
                                           record_on='update'),
                                RTDofGraph(name='Strain - Strain',
                                           var_x='U_k', idx_x=0,
                                           var_y='U_k', idx_y=1,
                                           record_on='update'),
                                RTDofGraph(name='Stress - Stress',
                                           var_x='F_int', idx_x=0,
                                           var_y='F_int', idx_y=1,
                                           record_on='update'),
                                RTDofGraph(name='sig1 - eps1',
                                           var_x='F_int', idx_x=0,
                                           var_y='U_k', idx_y=0,
                                           record_on='update'),
                                RTDofGraph(name='sig2 - sig3',
                                           var_x='F_int', idx_x=1,
                                           var_y='F_int', idx_y=2,
                                           record_on='update'),
                                RTDofGraph(name='eps2 - eps3',
                                           var_x='U_k', idx_x=1,
                                           var_y='U_k', idx_y=2,
                                           record_on='update')
                                ],
                'tline': TLine(step=0.1, max=1.0)
                }

        def _set_explorer_config(self, value):
            self._explorer_config = value



'''
Created on 15.02.2018

@author: abaktheer
'''

from .fets3D import FETS3D
import numpy as np
import sympy as sp
import traits.api as tr
from bmcs_utils.api import InjectSymbExpr, SymbExpr

xi_1, xi_2, xi_3 = sp.symbols('xi_1, xi_2, xi_3')

xi_i = sp.Matrix([xi_1, xi_2, xi_3])

#=================================================
# 8 nodes isoparametric volume element (3D)
#=================================================
import sympy as sp

class FETS3D8HSymbExpr(SymbExpr):

    xi_i = xi_i

    N_xi_i = sp.Matrix([(1 - xi_1) * (1 - xi_2) * (1 - xi_3) / 8,
                        (1 + xi_1) * (-1 + xi_2) * (-1 + xi_3) / 8,  # 2
                        (-1 + xi_1) * (1 + xi_2) * (-1 + xi_3) / 8,
                        (-1 - xi_1) * (-1 - xi_2) * (1 - xi_3) / 8,
                        (-1 + xi_1) * (-1 + xi_2) * (1 + xi_3) / 8,
                        (-1 - xi_1) * (1 - xi_2) * (-1 - xi_3) / 8,  # 6
                        (1 - xi_1) * (-1 - xi_2) * (-1 - xi_3) / 8,
                        (1 + xi_1) * (1 + xi_2) * (1 + xi_3) / 8, ]).T

    dN_xi_ai_ = [N_xi.diff(xi_i) for N_xi in N_xi_i]
    dN_xi_ai = sp.Matrix.hstack(*dN_xi_ai_)

    symb_model_params = []

    symb_expressions = [
        ('N_xi_i', ('xi_i',)),
        ('dN_xi_ai', ('xi_i',))
    ]


class FETS3D8H(FETS3D, InjectSymbExpr):

    symb_class = FETS3D8HSymbExpr

    dof_r = tr.Array(np.float64,
                     value=[[-1, -1, -1], [1, -1, -1],
                            [-1, 1, -1], [1, 1, -1],
                            [-1, -1, 1], [1, -1, 1],
                            [-1, 1, 1], [1, 1, 1], ])

    geo_r = tr.Array(np.float64,
                     value=[[-1, -1, -1], [1, -1, -1],
                            [-1, 1, -1], [1, 1, -1],
                            [-1, -1, 1], [1, -1, 1],
                            [-1, 1, 1], [1, 1, 1], ])
    vtk_r = tr.Array(np.float64,
                     value=[[-1, -1, -1], [1, -1, -1],
                            [-1, 1, -1], [1, 1, -1],
                            [-1, -1, 1], [1, -1, 1],
                            [-1, 1, 1], [1, 1, 1], ])
    n_nodal_dofs = 3

    delta = np.identity(n_nodal_dofs)

    vtk_cells = [[0, 1, 3, 2, 4, 5, 7, 6]]
    vtk_cell_types = 'Hexahedron'
    vtk_cell = [0, 1, 3, 2, 4, 5, 7, 6]
    vtk_cell_type = 'Hexahedron'

    vtk_expand_operator = tr.Array(np.float64, value=np.identity(3))

    # numerical integration points (IP) and weights
    xi_m = tr.Array(np.float64,
                    value=[[-1.0 / np.sqrt(3.0), -1.0 / np.sqrt(3.0), -1.0 / np.sqrt(3.0)],
                           [1.0 / np.sqrt(3.0), -1.0 /
                            np.sqrt(3.0), -1.0 / np.sqrt(3.0)],
                           [-1.0 / np.sqrt(3.0), 1.0 /
                            np.sqrt(3.0), -1.0 / np.sqrt(3.0)],
                           [1.0 / np.sqrt(3.0), 1.0 /
                            np.sqrt(3.0), -1.0 / np.sqrt(3.0)],
                           [-1.0 / np.sqrt(3.0), -1.0 /
                            np.sqrt(3.0), 1.0 / np.sqrt(3.0)],
                           [1.0 / np.sqrt(3.0), -1.0 /
                            np.sqrt(3.0), 1.0 / np.sqrt(3.0)],
                           [-1.0 / np.sqrt(3.0), 1.0 /
                            np.sqrt(3.0), 1.0 / np.sqrt(3.0)],
                           [1.0 / np.sqrt(3.0), 1.0 / np.sqrt(3.0),
                            1.0 / np.sqrt(3.0)],
                           ])

    w_m = tr.Array(value=[1, 1, 1, 1, 1, 1, 1, 1], dtype=np.float64)

    n_m = tr.Property

    def _get_n_m(self):
        return len(self.w_m)

    N_im = tr.Property()
    '''Shape function values in integration poindots.
    '''
    @tr.cached_property
    def _get_N_im(self):
        N_im = self.symb.get_N_xi_i(self.xi_m.T)
        return N_im

    dN_imr = tr.Property()
    '''Shape function derivatives in integration points.
    '''
    @tr.cached_property
    def _get_dN_imr(self):
        N_rim = self.symb.get_dN_xi_ai(self.xi_m.T)
        return np.einsum('rim->imr', N_rim)

    dN_inr = tr.Property()
    '''Shape function derivatives in visualization points.
    '''
    @tr.cached_property
    def _get_dN_inr(self):
        N_rin = self.symb.get_dN_xi_ai(self.dof_r.T)
        return np.einsum('rin->inr', N_rin)

    I_sym_abcd = tr.Array(np.float64)

    def _I_sym_abcd_default(self):
        delta = np.eye(3)
        return 0.5 * \
            (np.einsum('ac,bd->abcd', delta, delta) +
             np.einsum('ad,bc->abcd', delta, delta))

    plot_backend = 'k3d'

    def update_plot(self, k3d_plot):
        ax = axes
        import numpy as np
        v = np.linspace(-1,1,10)
        x, y, z = np.meshgrid(v,v,v)
        X_aIJK = np.array([x, y, z], dtype=np.float64)
        xmin, xmax, ymin, ymax, zmin, zmax = -1, 1, -1, 1, -1, 1
        N_IJK = self.symb.get_N_xi_i(X_aIJK)[0,...]
        N_IJK.shape
        import k3d
        plt_iso = k3d.marching_cubes(N_IJK[0,...],compression_level=9,
                                     xmin=xmin, xmax=xmax,
                                 ymin=ymin, ymax=ymax,
                                 zmin=zmin, zmax=zmax, level=0.5,
                                flat_shading=False)
        k3d_plot += plt_iso
        k3d_plot.display()



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Apr 20, 2010 by: rch

import sys

#--------------------------------------------------------------------------
# Helper methods specification
#--------------------------------------------------------------------------
def _find_class ( klass ):
    '''Helper method to import the specified class
    '''
    module = ''
    col    = klass.rfind( '.' )
    if col >= 0:
        module = klass[ : col ]
        klass = klass[ col + 1: ]

    mod = sys.modules.get( module )
    theClass = getattr( mod, klass, None )
    if (theClass is None) and (col >= 0):
        try:
            mod = __import__( module )
            for component in module.split( '.' )[1:]:
                mod = getattr( mod, component )

            theClass = getattr( mod, klass, None )
        except:
            pass

    return theClass
    



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Feb 16, 2010 by: rch

import numpy

def smooth(x,window_len=11,window='hanning'):
    """smooth the data using a window with requested size.
    
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    
    input:
        x: the input signal 
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal
        
    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)
    
    see also: 
    
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter
 
    TODO: the window parameter could be the window itself if an array instead of a string   
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")


    if window_len<3:
        return x


    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")


    s=numpy.r_[2*x[0]-x[window_len:1:-1],x,2*x[-1]-x[-1:-window_len:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w=ones(window_len,'d')
    else:
        w=eval('numpy.'+window+'(window_len)')

    y=numpy.convolve(w/w.sum(),s,mode='same')
    return y[window_len-1:-window_len+1]




from numpy import *
from pylab import *

def smooth_demo():

    t=linspace(-4,4,100)
    x=sin(t)
    xn=x+randn(len(t))*0.1
    y=smooth(x)

    ws=31

    subplot(211)
    plot(ones(ws))

    windows=['flat', 'hanning', 'hamming', 'bartlett', 'blackman']

    hold(True)
    for w in windows[1:]:
        eval('plot('+w+'(ws) )')

    axis([0,30,0,1.1])

    legend(windows)
    title("The smoothing windows")
    subplot(212)
    plot(x)
    plot(xn)
    for w in windows:
        plot(smooth(xn,10,w))
    l=['original signal', 'signal with noise']
    l.extend(windows)

    legend(l)
    title("Smoothing a noisy signal")
    show()


if __name__=='__main__':
    smooth_demo()



from ibvpy.mesh.i_fe_grid_slice import IFENodeSlice
from ibvpy.mathkit.linalg.sys_mtx_assembly import SysMtxArray
from ibvpy.mathkit.tensor import EPS
from ibvpy.sim.i_xmodel import IXModel
from traits.api import \
    provides, Tuple, \
    Array, Property, Instance, cached_property

import numpy as np

from .xdomain_transform import XDomainFEGridTransform


@provides(IXModel)
class XDomainFEInterface(XDomainFEGridTransform):

    #=========================================================================
    # CRIMINAL - Change
    #=========================================================================
    state_var_shape = Property(Tuple)

    def _get_state_var_shape(self):
        n_elems = len(self.o_Epia)
        return (n_elems, self.fets.n_m,)

    vtk_expand_operator = Array(np.float64)

    def _vtk_expand_operator_default(self):
        return np.identity(3)

    DELTA_p = Array()

    def _DELTA_p_default(self):
        return np.array([-1, 1], np.float64)

    I = Instance(IFENodeSlice)
    J = Instance(IFENodeSlice)

    o_Epia = Property(depends_on='changed')

    @cached_property
    def _get_o_Epia(self):
        o_Ia = self.I.dofs
        o_Ja = self.J.dofs
        o_piEa = np.array([[o_Ia[:-1], o_Ia[1:]],
                           [o_Ja[:-1], o_Ja[1:]]])
        return np.einsum('piEa->Epia', o_piEa)

    o_Eia = Property(depends_on='changed')
    '''flatten the first two dimensions consistently woth o_pEia
    '''
    @cached_property
    def _get_o_Eia(self):
        flat_bnd = (-1,) + self.o_Epia.shape[2:]
        return self.o_Epia.reshape(flat_bnd)

    X_pEia = Property(depends_on='changed')

    @cached_property
    def _get_X_pEia(self):
        X_Ia = self.I.geo_X
        X_Ja = self.J.geo_X
        X_piEa = np.array([[X_Ia[:-1], X_Ia[1:]],
                           [X_Ja[:-1], X_Ja[1:]]])
        return np.einsum('piEa->pEia', X_piEa)

    x_Eia = Property(depends_on='changed')

    @cached_property
    def _get_x_Eia(self):
        DELTA_pp = np.identity(2)
        # Correct the FETS interface - to provide shape functions
        # at integration points.
        return np.einsum(
            'pp,mi,pEma->Eia',
            DELTA_pp, self.fets.N_mi_geo, self.X_pEia
        ) / 2.0

    det_J_Em = Property(depends_on='MESH,GEO,CS,FE')
    '''Get the determinant of the local jacobi matrix
    '''
    @cached_property
    def _get_det_J_Em(self):
        r_Eia = np.einsum(
            'Eira,Eia->Eir',
            self.T_Emra[..., :self.x_Eia.shape[-1]], self.x_Eia
        )
        J_Emar = np.einsum(
            'imr,Eia->Emar', self.fets.dN_imr, r_Eia[..., :1]
        )
        return np.linalg.det(J_Emar)

    T_Emra = Property(depends_on='MESH,GEO,CS,FE')

    @cached_property
    def _get_T_Emra(self):

        x_Eia = self.x_Eia
        dx_Emas = np.einsum(
            'ims,Eia->Emas', self.fets.dN_imr, x_Eia
        )
        DELTA_12 = np.array([[1, 0]], np.float64)
        dx2_Emar = np.einsum('sr,Emas->Emar',
                             DELTA_12, dx_Emas
                             )
        # expansion tensor
        DELTA23_ab = np.array([[1, 0, 0],
                               [0, 1, 0]], dtype=np.float64)

        dx3_Emar = np.einsum(
            'ik,...ij,jl->...kl',
            DELTA23_ab, dx2_Emar, DELTA23_ab
        )
        dx3_Emar[..., 2, 2] = 1.0

        m_0_Ema = dx3_Emar[..., 0]

        m_1_Ema = np.einsum(
            '...i,...j,ijk->...k',
            dx3_Emar[..., 2], m_0_Ema, EPS
        )
        m_2_Ema = np.einsum(
            '...i,...j,ijk->...k',
            m_0_Ema, m_1_Ema, EPS)
        M_rEma = np.array([m_0_Ema, m_1_Ema, m_2_Ema])
        M_Emra = np.einsum('rEma->Emra', M_rEma)
        norm_M_Emra = np.sqrt(
            np.einsum('...ij,...ij->...i', M_Emra, M_Emra)
        )[..., np.newaxis]

        T_Emra = M_Emra / norm_M_Emra
        return T_Emra

    B_Eimabc = Property(depends_on='MESH,GEO,CS,FE')
    '''Kinematic mapping between displacements and strains in every
    integration point.
    '''
    @cached_property
    def _get_B_Eimabc(self):
        return np.einsum(
            'p,Emra,im->Empira',
            self.DELTA_p, self.T_Emra[..., :2, :2], self.fets.N_im
        )

    def _get_BB_Emicjdabef(self):
        return np.einsum(
            '...Empira,...Emqjsb, Em, m->...Empirqjsab',
            self.B_Eimabc, self.B_Eimabc, self.det_J_Em, self.fets.w_m
        )

    n_dofs = Property

    def _get_n_dofs(self):
        return 0

    def map_U_to_field(self, U):
        U_Epia = U[self.o_Epia]
        s_Emr = np.einsum(
            'Empira,Epia->Emr',
            self.B_Eimabc, U_Epia
        )
        return s_Emr

    def map_field_to_K(self, D_Emabef):
        K_Eicjd = self.integ_factor * np.einsum(
            '...Empirqjsab,...Emrs->Epiaqjb',
            self.BB_Emicjdabef, D_Emabef
        )
        _, _, n_p, n_i, _, n_a = self.B_Eimabc.shape
        n_o = n_p * n_i * n_a
        K_Eij = K_Eicjd.reshape(-1, n_o, n_o)
        o_Ei = self.o_Epia.reshape(-1, n_o)
        return SysMtxArray(mtx_arr=K_Eij, dof_map_arr=o_Ei)

    def map_field_to_F(self, sig_Emab):
        _, _, n_p, n_i, _, n_a = self.B_Eimabc.shape
        f_Epia = self.integ_factor * np.einsum(
            'm,Empira,Emr,Em->Epia',
            self.fets.w_m, self.B_Eimabc, sig_Emab, self.det_J_Em
        )
        n_o = n_p * n_i * n_a
        f_Ei = f_Epia.reshape(-1, n_o)
        o_E = self.o_Eia.reshape(-1, n_o)
        return o_E.flatten(), f_Ei.flatten()





from traits.etsconfig.api import ETSConfig

if ETSConfig.toolkit == 'wx':
    from .mpl_figure_editor_wx import MPLFigureEditor
elif ETSConfig.toolkit == 'qt4':
    from .mpl_figure_editor_qt import MPLFigureEditor
else:
    from .mpl_figure_editor_void import MPLFigureEditor
#     raise ImportError("MPLEditor for %s toolkit not available" %
#                       ETSConfig.toolkit)
